# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore


import math

from docplex.mp.compat23 import izip
from docplex.mp.constr import AbstractConstraint, LinearConstraint,\
    LogicalConstraint, EquivalenceConstraint, IndicatorConstraint, QuadraticConstraint
from docplex.mp.error_handler import docplex_fatal
from docplex.mp.linear import Var, Expr
from docplex.mp.pwl import PwlFunction
from docplex.mp.progress import ProgressListener
from docplex.mp.utils import is_int, is_number, is_iterable, is_string, generate_constant, \
    is_ordered_sequence, is_iterator, resolve_caller_as_string
from docplex.mp.vartype import VarType
import six

_vartype_code_map = {sc().get_cplex_typecode(): sc().short_name for sc in VarType.__subclasses__()}


def vartype_code_to_string(vartype_code):
    return _vartype_code_map.get(vartype_code, '????')


class DocplexNumericCheckerMixin(object):

    @staticmethod
    def static_validate_num1(e, checked_num=False, infinity=1e+20):
        # checks for number and truncates to 1e=20
        if not checked_num and not is_number(e):
            docplex_fatal("Expecting number, got: {0!r}".format(e))
        elif -infinity <= e <= infinity:
            return e
        elif e >= infinity:
            return infinity
        else:
            return -infinity

    @staticmethod
    def static_validate_num2(e, infinity=1e+20, context_msg=None):
        # checks for number, nans, nath.inf, and truncates to 1e+20
        if not is_number(e):
            docplex_fatal("Not a number: {}".format(e))
        elif math.isnan(e):
            msg = "NaN value found in expression"
            if context_msg is not None:
                try:
                    msg = "{0}: {1}".format(context_msg(), msg)
                except TypeError:
                    msg = "{0}: {1}".format(context_msg, msg)
            docplex_fatal(msg)
        elif math.isinf(e):
            msg = "Infinite value forbidden in expression"
            if context_msg is not None:
                try:
                    msg = "{0}: {1}".format(context_msg(), msg)
                except TypeError:
                    msg = "{0}: {1}".format(context_msg, msg)
            docplex_fatal(msg)
        elif -infinity <= e <= infinity:
            return e
        elif e >= infinity:
            return infinity
        else:
            return -infinity

    @classmethod
    def typecheck_num_seq(cls, logger, seq, check_math, caller=None):
        # build a list to avoid consuming an iterator
        checked_num_list = list(seq)
        for i, x in enumerate(checked_num_list):
            def loop_caller():
                return "%s, pos %d" % (caller, i) if caller else ""

            cls.typecheck_num(logger, x, check_math, loop_caller)
        return checked_num_list

    @classmethod
    def typecheck_num(cls, logger, arg, check_math, caller=None):
        if not is_number(arg):
            caller_string = resolve_caller_as_string(caller)
            logger.fatal("{0}Expecting number, got: {1!r}", (caller_string, arg))
        elif check_math:
            if math.isnan(arg):
                caller_string = resolve_caller_as_string(caller)
                logger.fatal("{0}NaN value detected", (caller_string,))
            elif math.isinf(arg):
                caller_string = resolve_caller_as_string(caller)
                logger.fatal("{0}Infinite value detected", (caller_string,))

    @classmethod
    def typecheck_int(cls, logger, arg, check_math, accept_negative=True, caller=None):
        if not is_number(arg):
            caller_string = resolve_caller_as_string(caller)
            logger.fatal("{0}Expecting number, got: {1!r}", (caller_string, arg))
        if check_math:
            if math.isnan(arg):
                caller_string = resolve_caller_as_string(caller)
                logger.fatal("{0}NaN value detected", (caller_string,))
            elif math.isinf(arg):
                caller_string = resolve_caller_as_string(caller)
                logger.fatal("{0}Infinite value detected", (caller_string,))
        if not is_int(arg):
            caller_string = resolve_caller_as_string(caller)
            logger.fatal("{0}Expecting integer, got: {1!r}", (caller_string, arg))
        elif not accept_negative and arg < 0:
            caller_string = resolve_caller_as_string(caller)
            logger.fatal("{0}Expecting positive integer, got: {1!r}", (caller_string, arg))


class DocplexTypeCheckerI(object):

    def check_sync_indices(self):
        return True

    def typecheck_iterable(self, arg):
        raise NotImplementedError  # pragma: no cover

    def typecheck_valid_index(self, arg):
        raise NotImplementedError  # pragma: no cover

    def typecheck_vartype(self, arg):
        raise NotImplementedError  # pragma: no cover

    def typecheck_var(self, obj, vartype=None):
        raise NotImplementedError  # pragma: no cover

    def typecheck_binary_var(self, obj):
        return self.typecheck_var(obj, vartype='B')

    def typecheck_continuous_var(self, obj):
        return self.typecheck_var(obj, vartype='C')

    def typecheck_var_seq(self, seq, vtype=None, caller=None):
        return seq  # pragma: no cover

    def typecheck_logical_op_seq(self, seq):
        return seq  # pragma: no cover

    def typecheck_logical_op(self, arg, caller):
        raise NotImplementedError  # pragma: no cover

    def typecheck_var_seq_all_different(self, seq):
        raise NotImplementedError  # pragma: no cover

    def typecheck_num_seq(self, seq, caller=None):
        raise NotImplementedError  # pragma: no cover

    def typecheck_operand(self, obj, accept_numbers=True, caller=None):
        raise NotImplementedError  # pragma: no cover

    def typecheck_constraint(self, obj):
        raise NotImplementedError  # pragma: no cover

    def typecheck_ct_to_add(self, ct, mdl, caller):
        raise NotImplementedError  # pragma: no cover

    def typecheck_ct_not_added(self, ct, do_raise=False, caller=None):
        raise NotImplementedError  # pragma: no cover

    def typecheck_cts_added_to_model(self, mdl, cts, caller=None):
        return cts  # pragma: no cover

    def typecheck_linear_constraint(self, obj, accept_ranges=True):
        raise NotImplementedError  # pragma: no cover

    def typecheck_constraint_seq(self, cts, check_linear=False, accept_range=True):
        # must return sequence unchanged
        return cts  # pragma: no cover

    def typecheck_logical_constraint_seq(self, cts, true_if_equivalence):
        # must return sequence unchanged
        return cts  # pragma: no cover

    def typecheck_quadratic_constraint_seq(self, cts):
        # must return sequence unchanged
        return cts  # pragma: no cover

    def typecheck_linear_constraint_name_tuple_seq(self, ct_ctname_seq, accept_range=True):
        # must return sequence unchanged
        return ct_ctname_seq  # pragma: no cover

    def typecheck_zero_or_one(self, arg):
        raise NotImplementedError  # pragma: no cover

    def typecheck_num(self, arg, caller=None):
        raise NotImplementedError  # pragma: no cover

    def typecheck_int(self, arg, accept_negative=False, caller=None):
        raise NotImplementedError  # pragma: no cover

    def check_vars_domain(self, lbs, ubs, names):
        raise NotImplementedError  # pragma: no cover

    def check_var_domain(self, lbs, ubs, names):
        raise NotImplementedError  # pragma: no cover

    def typecheck_string(self, arg, accept_empty=False, accept_none=False, caller=''):
        raise NotImplementedError  # pragma: no cover

    def typecheck_string_seq(self, arg, accept_empty=False, accept_none=False, caller=''):
        raise NotImplementedError  # pragma: no cover

    def typecheck_in_model(self, model, mobj, caller=''):
        raise NotImplementedError  # pragma: no cover

    def typecheck_key_seq(self, keys, accept_empty_seq=False):
        raise NotImplementedError  # pragma: no cover

    def get_number_validation_fn(self):
        raise NotImplementedError  # pragma: no cover

    def typecheck_progress_listener(self, arg):
        raise NotImplementedError  # pragma: no cover

    def typecheck_two_in_model(self, model, obj1, obj2, ctx_msg):
        raise NotImplementedError  # pragma: no cover

    def check_ordered_sequence(self, arg, caller, accept_iterator=True):
        raise NotImplementedError  # pragma: no cover

    def check_trivial_constraints(self):
        raise NotImplementedError  # pragma: no cover

    def check_solution_hook(self, mdl, sol_hook_fn):
        raise NotImplementedError

    def typecheck_pwl_function(self, pwl):
        raise NotImplementedError

    def check_duplicate_name(self, name, name_table, qualifier):
        raise NotImplementedError

    def check_for_duplicate_keys(self, keys, caller=None):
        # default is no-op
        pass


# noinspection PyAbstractClass
class DOcplexLoggerTypeChecker(DocplexTypeCheckerI):
    def __init__(self, logger):
        self._logger = logger

    def fatal(self, msg, *args):
        self._logger.fatal(msg, args)

    def error(self, msg, *args):  # pragma: no cover
        self._logger.error(msg, args)

    def warning(self, msg, *args):  # pragma: no cover
        self._logger.warning(msg, args)


class StandardTypeChecker(DOcplexLoggerTypeChecker):

    def __init__(self, logger):
        DOcplexLoggerTypeChecker.__init__(self, logger)

    @property
    def name(self):
        return "std"

    def typecheck_iterable(self, arg):
        # INTERNAL: checks for an iterable
        if not is_iterable(arg):
            self.fatal("Expecting iterable, got: {0!s}", arg)

    # safe checks.
    def typecheck_valid_index(self, arg):
        if arg < 0:
            self.fatal("Invalid index: {0!s}", arg)

    def typecheck_vartype(self, arg):
        # INTERNAL: check for a valid vartype
        if not isinstance(arg, VarType):
            self.fatal("Not a variable type: {0!s}, type: {1!s}", arg, type(arg))
        return True

    def typecheck_var(self, obj, vartype=None):
        # INTERNAL: check for Var instance
        if not isinstance(obj, Var):
            self.fatal("Expecting decision variable, got: {0!s} type: {1!s}", obj, type(obj))
        if vartype and obj.vartype.get_cplex_typecode() != vartype:
            self.fatal("Expecting {0} variable, got: {1!s} type: {2!s}",
                       vartype_code_to_string(vartype), obj, obj.vartype)

    def typecheck_var_seq(self, seq, vtype=None, caller=None):
        # build a list to avoid consuming an iterator
        checked_var_list = list(seq)
        for i, x in enumerate(checked_var_list):
            if not isinstance(x, Var):
                caller_s = resolve_caller_as_string(caller)
                self.fatal("{2}Expecting an iterable returning variables, {0!r} was passed at position {1}", x, i, caller_s)
            if vtype and x.vartype.get_cplex_typecode() != vtype:
                caller_s = resolve_caller_as_string(caller)
                self.fatal("{3}Expecting an iterable returning variables of type {0}, {1!r} was passed at position {2}",
                           vtype.short_name, x, i, caller_s)

        return checked_var_list

    def typecheck_logical_op_seq(self, seq, caller=None):
        checked_args = list(seq)
        for i, x in enumerate(checked_args):
            if caller is None:
                loop_caller = None
            else:
                def loop_caller():
                    return '%s, arg#%d' % (resolve_caller_as_string(caller, sep=''), i)
            self.typecheck_logical_op(x, caller=loop_caller)
        return checked_args

    def typecheck_logical_op(self, arg, caller):
        if not hasattr(arg, 'as_logical_operand') or arg.as_logical_operand() is None:
            caller_s = resolve_caller_as_string(caller)
            self.fatal('{1}Not a logical operand: {0!r}. Expecting binary variable, logical expression', arg, caller_s)

    def typecheck_num_seq(self, seq, caller=None):
        return DocplexNumericCheckerMixin.typecheck_num_seq(self._logger, seq, check_math=False, caller=caller)

    def typecheck_var_seq_all_different(self, seq):
        # return the checked sequence, so take the list
        seq_as_list = list(seq)
        for v in seq_as_list:
            self.typecheck_var(v)
        # check for all differemt and output a justifier variable apperaing twice.
        inc_set = set([])
        for v in seq_as_list:
            if v.index in inc_set:
                self.fatal('Variable: {0} appears twice in sequence', v)

            else:
                inc_set.add(v.index)
        return seq_as_list

    def typecheck_constraint(self, obj):
        if not isinstance(obj, AbstractConstraint):
            self.fatal("Expecting constraint, got: {0!s} with type: {1!s}", obj, type(obj))

    def typecheck_ct_to_add(self, ct, mdl, caller):
        if not isinstance(ct, AbstractConstraint):
            self.fatal("Expecting constraint, got: {0!r} with type: {1!s}", ct, type(ct))
        self.typecheck_in_model(mdl, ct, caller)

    def typecheck_ct_not_added(self, ct, do_raise=False, caller=None):
        if ct.is_added():
            s_caller = resolve_caller_as_string(caller, sep=' ')
            if do_raise:
                self.fatal('{0}expects a non-added constraint, {1} is added (index={2})',
                           s_caller, ct, ct.get_index()
                           )
            else:
                self.warning('{0}expects a non-added constraint, {1} is added (index={2})',
                             s_caller, ct, ct.get_index()
                             )

    def typecheck_cts_added_to_model(self, mdl, cts, caller=None):
        lcts = list(cts)
        for ct in lcts:
            if not ct.is_added():
                s_caller = resolve_caller_as_string(caller, sep=' ')
                mdl.fatal("{0}Constraint: {1!s} has not been added to any model".format(s_caller, ct))
            elif mdl is not ct.model:
                s_caller = resolve_caller_as_string(caller, sep=' ')
                mdl.fatal("{0}Constraint: {1!s} belongs to a different model".format(s_caller, ct))
        return lcts

    def typecheck_ct_added_to_model(self, mdl, ct, caller=None):
        if not ct.is_added():
            s_caller = resolve_caller_as_string(caller, sep=' ')
            mdl.fatal("{0}Constraint: {1!s} has not been added to any model".format(s_caller, ct))
        elif mdl is not ct.model:
            s_caller = resolve_caller_as_string(caller, sep=' ')
            mdl.fatal("{0}Constraint: {1!s} belongs to a different model".format(s_caller, ct))

    def typecheck_linear_constraint(self, obj, accept_range=True):
        if accept_range:
            if not isinstance(obj, AbstractConstraint):
                self.fatal("Expecting linear constraint, got: {0!r}", obj)
            if not obj.is_linear():
                self.fatal("Expecting linear constraint, got: {0!s} with type: {1!s}", obj, type(obj))
        else:
            if not isinstance(obj, LinearConstraint):
                self.fatal("Expecting linear constraint, got: {0!s} with type: {1!s}", obj, type(obj))

    def typecheck_constraint_seq(self, cts, check_linear=False, accept_range=True):
        checked_cts_list = list(cts)
        for i, ct in enumerate(checked_cts_list):
            if not isinstance(ct, AbstractConstraint):
                self.fatal("Expecting sequence of constraints, got: {0!r} at position {1}", ct, i)
            if check_linear:
                if not ct.is_linear():
                    self.fatal("Expecting sequence of linear constraints, got: {0!r} at position {1}", ct, i)
                elif not accept_range and not isinstance(ct, LinearConstraint):
                    self.fatal("Expecting sequence of linear constraints (not ranges), got: {0!r} at position {1}", ct,
                               i)
        return checked_cts_list

    def typecheck_logical_constraint_seq(self, cts, true_if_equivalence):
        checked_cts_list = list(cts)  # listify to avoid consuming iterators....
        if true_if_equivalence:
            checked_type = EquivalenceConstraint
            typename = "equivalence"
        elif true_if_equivalence is False:
            checked_type = IndicatorConstraint
            typename = "indicator"
        else:
            checked_type = LogicalConstraint
            typename = "equivalence or indicator"
        for i, ct in enumerate(checked_cts_list):
            if not isinstance(ct, checked_type):
                self.fatal("Expecting sequence of {0} constraints, got: {1!r} at position {2}", typename, ct, i)

        return checked_cts_list

    def typecheck_quadratic_constraint_seq(self, cts):
        checked_qcts_list = list(cts)  # listify to avoid consuming iterators....
        for i, ct in enumerate(checked_qcts_list):
            if not isinstance(ct, QuadraticConstraint):
                self.fatal("Expecting sequence of quadratic constraints, got: {0!r} at position {1}", ct, i)

        return checked_qcts_list

    def typecheck_linear_constraint_name_tuple_seq(self, ct_ctname_seq, accept_range=True):
        # must return sequence unchanged
        checked_list = list(ct_ctname_seq)
        for c, (ct, ctname) in enumerate(ct_ctname_seq):
            self.typecheck_linear_constraint(ct, accept_range=accept_range)
            # noinspection PyArgumentEqualDefault
            self.typecheck_string(ctname, accept_empty=True, accept_none=False)

        return checked_list

    def typecheck_zero_or_one(self, arg):
        if arg != 0 and arg != 1:
            self.fatal("expecting 0 or 1, got: {0!s}", arg)

    def typecheck_num(self, arg, caller=None):
        if not is_number(arg):
            caller_string = "{0}: ".format(caller) if caller is not None else ""
            self.fatal("{0}Expecting number, got: {1!r}", caller_string, arg)

    def typecheck_int(self, arg, accept_negative=True, caller=None):
        caller_string = "{0}: ".format(caller) if caller is not None else ""
        if not is_number(arg):
            self.fatal("{0}Expecting number, got: {1!r}", caller_string, arg)
        elif not is_int(arg):
            self.fatal("{0}Expecting integer, got: {1!r}", caller_string, arg)
        elif not accept_negative and arg < 0:
            self.fatal("{0}Expecting positive integer, got: {1!r}", caller_string, arg)

    def check_vars_domain(self, lbs, ubs, names):
        l_ubs = len(ubs)
        l_lbs = len(lbs)
        if l_lbs and l_ubs:
            names = names or generate_constant(None, max(l_lbs, l_ubs))
            # noinspection PyArgumentList,PyArgumentList
            for lb, ub, varname in izip(lbs, ubs, names):
                self.check_var_domain(lb, ub, varname)

    def check_var_domain(self, lb, ub, varname):
        if lb is not None and ub is not None and lb > ub:
            self.fatal('Empty variable domain, name={0}, lb={1}, ub={2}'.format(varname, lb, ub))

    def typecheck_string(self, arg, accept_empty=False, accept_none=False, caller=''):
        if is_string(arg):
            if not accept_empty and 0 == len(arg):
                s_caller = resolve_caller_as_string(caller)
                self.fatal("{0}Expecting a non-empty string", s_caller)
        elif not (arg is None and accept_none):
            s_caller = resolve_caller_as_string(caller)
            self.fatal("{0}Expecting string, got: {1!r}", s_caller, arg)

    def typecheck_string_seq(self, arg, accept_empty=False, accept_none=False, caller=''):
        checked_strings = list(arg)
        # do not accept a string
        if is_string(arg):
            s_caller = resolve_caller_as_string(caller)
            self.fatal("{0}Expecting list of strings, a string was passed: '{1}'", s_caller, arg)
        for s in checked_strings:
            self.typecheck_string(s, accept_empty=accept_empty, accept_none=accept_none, caller=caller)
        return checked_strings

    def typecheck_in_model(self, model, mobj, caller=''):
        # produces message of the type: "constraint ... does not belong to model
        if mobj.model != model:
            self.fatal("{0} ({2!s}) is not in model '{1:s}'".format(caller, model.name, mobj))

    def typecheck_key_seq(self, keys, accept_empty_seq=False):
        if any(k is None for k in keys):
            self.fatal("Variable keys cannot be None, got: {0!r}", keys)

    def get_number_validation_fn(self):
        return DocplexNumericCheckerMixin.static_validate_num1

    @staticmethod
    def _is_operand(arg, accept_numbers=True):
        return isinstance(arg, (Expr, Var)) or (accept_numbers and is_number(arg))

    def typecheck_operand(self, arg, accept_numbers=True, caller=None):
        if not self._is_operand(arg, accept_numbers=accept_numbers):
            caller_str = "{0}: ".format(caller) if caller else ""
            accept_str = "Expecting expr/var"
            if accept_numbers:
                accept_str += "/number"
            self.fatal("{0}{1}, got: {2!r}", caller_str, accept_str, arg)

    def typecheck_progress_listener(self, arg):
        if not isinstance(arg, ProgressListener):
            self.fatal('Expecting ProgressListener instance, got: {0!r}', arg)

    def typecheck_two_in_model(self, model, mobj1, mobj2, ctx_msg):
        mobj1_model = mobj1._get_model()
        mobj2_model = mobj2._get_model()
        if mobj1_model != mobj2_model:
            self.fatal("Cannot mix objects from different models in {0}. obj1={1!s}, obj2={2!s}"
                       .format(ctx_msg, mobj1, mobj2))
        elif mobj1_model != model:
            self.fatal("Objects do not belong to model {0}. obj1={1!s}, obj2={2!s}"
                       .format(self, mobj1, mobj2))

    def check_trivial_constraints(self):
        return True

    def check_ordered_sequence(self, arg, caller, accept_iterator=True):
        # in some cases, we need an ordered sequence, if not the code won't crash
        # but may do unexpected things
        if not(is_ordered_sequence(arg) or (accept_iterator and is_iterator(arg))):
            self.fatal("{0}, got: {1!s}", caller, type(arg).__name__)

    def check_solution_hook(self, mdl, sol_hook_fn):
        if not callable(sol_hook_fn):
            self.fatal('Solution hook requires a function taking a solution as argument, a non-callable was passed')
        if six.PY3:
            try:
                from inspect import signature
                hook_signature = signature(sol_hook_fn)
                nb_params = len(hook_signature.parameters)
                if nb_params != 1:
                    self.fatal(
                        'Solution hook requires a function taking a solution as argument, wrong number of arguments: {0}'
                        .format(nb_params))
            except (ImportError, TypeError):  # not a callable object or no signature
                pass

    def typecheck_pwl_function(self, pwl):
        if not isinstance(pwl, PwlFunction):
            self.fatal('Expecting piecewise-linear function, {0!r} was passed', pwl)

    def check_duplicate_name(self, name, name_table, qualifier):
        if name_table is not None:
            if name in name_table:
                self.warning("Duplicate {2} name: {0!s}, used for: {1}", name, name_table[name], qualifier)


class DummyTypeChecker(DOcplexLoggerTypeChecker):

    def check_sync_indices(self):
        return False

    # noinspection PyUnusedLocal
    def __init__(self, logger):
        super(DummyTypeChecker, self).__init__(logger)

    @property
    def name(self):
        return "off"

    def typecheck_iterable(self, arg):
        pass  # pragma: no cover

    def typecheck_valid_index(self, arg):
        pass  # pragma: no cover

    def typecheck_vartype(self, arg):
        pass  # pragma: no cover

    def typecheck_var(self, obj, vartype=None):
        pass  # pragma: no cover

    def typecheck_var_seq(self, seq, vtype=None, caller=None):
        return seq  # pragma: no cover

    def typecheck_num_seq(self, seq, caller=None):
        return seq  # pragma: no cover

    def typecheck_var_seq_all_different(self, seq):
        return seq

    def typecheck_operand(self, obj, accept_numbers=True, caller=None):
        pass  # pragma: no cover

    def typecheck_constraint(self, obj):
        pass  # pragma: no cover

    def typecheck_ct_to_add(self, ct, mdl, caller):
        pass  # pragma: no cover

    def typecheck_ct_not_added(self, ct, do_raise=False, caller=None):
        pass

    def typecheck_cts_added_to_model(self, mdl, cts, caller=None):
        return cts

    def typecheck_ct_added_to_model(self, mdl, ct, caller=None):
        pass

    def typecheck_linear_constraint(self, obj, accept_range=True):
        pass  # pragma: no cover

    def typecheck_constraint_seq(self, cts, check_linear=False, accept_range=True):
        # must return sequence unchanged
        return cts  # pragma: no cover

    def typecheck_logical_constraint_seq(self, cts, true_if_equivalence):
        # must return sequence unchanged
        return cts  # pragma: no cover

    def typecheck_quadratic_constraint_seq(self, cts):
        # must return sequence unchanged
        return cts  # pragma: no cover

    def typecheck_linear_constraint_name_tuple_seq(self, ct_ctname_seq, accept_range=True):
        # must return sequence unchanged
        return ct_ctname_seq  # pragma: no cover

    def typecheck_zero_or_one(self, arg):
        pass  # pragma: no cover

    def typecheck_num(self, arg, caller=None):
        pass  # pragma: no cover

    def typecheck_int(self, arg, accept_negative=True, caller=None):
        pass  # pragma: no cover

    def check_vars_domain(self, lbs, ubs, names):
        # do nothing on variable bounds
        pass

    def check_var_domain(self, lb, ub, varname):
        pass

    def typecheck_string(self, arg, accept_empty=False, accept_none=False, caller=''):
        pass  # pragma: no cover

    def typecheck_string_seq(self, arg, accept_empty=False, accept_none=False, caller=''):
        return arg

    def typecheck_in_model(self, model, mobj, caller=''):
        pass  # pragma: no cover

    def typecheck_key_seq(self, keys, accept_empty_seq=False):
        pass  # pragma: no cover

    def typecheck_progress_listener(self, arg):
        pass  # pragma: no cover

    def typecheck_two_in_model(self, model, obj1, obj2, ctx_msg):
        pass  # pragma: no cover

    def check_ordered_sequence(self, arg, caller, accept_iterator=True):
        pass  # pragma: no cover

    def check_trivial_constraints(self):
        return False

    def get_number_validation_fn(self):
        return None

    def check_solution_hook(self, mdl, sol_hook_fn):
        pass

    def typecheck_pwl_function(self, pwl):
        pass

    def check_duplicate_name(self, name, name_table, qualifier):
        pass

    def typecheck_logical_op(self, arg, caller):
        pass


class NumericTypeChecker(DummyTypeChecker):

    def __init__(self, logger):
        super(NumericTypeChecker, self).__init__(logger)

    @property
    def name(self):
        return "numeric"

    def get_number_validation_fn(self):
        return DocplexNumericCheckerMixin.static_validate_num2

    def typecheck_num(self, arg, caller=None):
        DocplexNumericCheckerMixin.typecheck_num(self._logger, arg, check_math=True, caller=caller)

    def typecheck_int(self, arg, accept_negative=True, caller=None):
        DocplexNumericCheckerMixin.typecheck_int(self._logger, arg, check_math=True, accept_negative=accept_negative,
                                                 caller=caller)

    def typecheck_num_seq(self, seq, caller=None):
        return DocplexNumericCheckerMixin.typecheck_num_seq(self._logger, seq, check_math=True, caller=caller)


class FullTypeChecker(StandardTypeChecker):

    def __init__(self, logger):
        super(FullTypeChecker, self).__init__(logger)

    @property
    def name(self):
        return "full"

    def get_number_validation_fn(self):
        return DocplexNumericCheckerMixin.static_validate_num2

    def typecheck_num(self, arg, caller=None):
        DocplexNumericCheckerMixin.typecheck_num(self._logger, arg, check_math=True, caller=caller)

    def typecheck_int(self, arg, accept_negative=True, caller=None):
        DocplexNumericCheckerMixin.typecheck_int(self._logger, arg, check_math=True, accept_negative=accept_negative,
                                                 caller=caller)

    def typecheck_num_seq(self, seq, caller=None):
        return DocplexNumericCheckerMixin.typecheck_num_seq(self._logger, seq, check_math=True, caller=caller)

    def check_for_duplicate_keys(self, keys, caller=None):
        key_set = set(keys)
        if len(key_set) < len(keys):
            # some key is duplicated:
            inc_set = set()
            for k in keys:
                if k in inc_set:
                    s_caller = resolve_caller_as_string(caller, sep=' ')
                    self.fatal("{0}Duplicated key: {1!s}".format(s_caller, k))
                else:
                    inc_set.add(k)


#  ------------------------------
# noinspection PyPep8
_tck_map = {'default': StandardTypeChecker,
            'standard': StandardTypeChecker,
            'std': StandardTypeChecker,
            'on': StandardTypeChecker,
            # --
            'numeric': NumericTypeChecker,
            'full': FullTypeChecker,
            # --
            'off': DummyTypeChecker,
            'deploy': DummyTypeChecker,
            'no_checks': DummyTypeChecker}


def get_typechecker(arg, logger):
    if arg:
        key = arg.lower()
        if key in _tck_map:
            checker_type = _tck_map[key]
        else:
            msg = 'Unexpected typechecker key: {0} - expecting on|off|std|default|numeric|full. Using default'.format(
                key)
            if logger:
                logger.error(msg)
            else:
                print('*Warning: {0}'.format(msg))
            checker_type = StandardTypeChecker

    else:
        checker_type = StandardTypeChecker
    return checker_type(logger)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore

from docplex.mp.sosvarset import SOSVariableSet
from docplex.mp.operand import LinearOperand
from docplex.mp.linear import Var, LinearExpr, AbstractLinearExpr, ZeroExpr, ConstantExpr
from docplex.mp.operand import Operand
from docplex.mp.constants import ComparisonType, UpdateEvent, ObjectiveSense
from docplex.mp.constr import LinearConstraint, RangeConstraint, \
    IndicatorConstraint, PwlConstraint, EquivalenceConstraint, IfThenConstraint, NotEqualConstraint, QuadraticConstraint
from docplex.mp.functional import MaximumExpr, MinimumExpr, AbsExpr, PwlExpr, LogicalAndExpr, LogicalOrExpr, \
    LogicalNotExpr
from docplex.mp.pwl import PwlFunction
from docplex.mp.compat23 import fast_range, izip_longest
from docplex.mp.environment import Environment
from docplex.mp.utils import *
from docplex.mp.kpi import KPI
from docplex.mp.solution import SolveSolution
from docplex.mp.sttck import StaticTypeChecker

from itertools import product, islice


class ModificationBlocker(object):
    fmt = "Cplex cannot modify {0} in-place"

    all_blockers = {}

    def __init__(self, cause):
        self._msg = self.fmt.format(cause)

    def cannot_modify(self, expr):
        raise DOcplexException("{0}: {1}".format(self._msg, expr))

    def notify_expr_modified(self, expr, event):  # event is ignored
        self.cannot_modify(expr)

    @classmethod
    def get_blocker(cls, name):
        cached_blocker = cls.all_blockers.get(name)
        if cached_blocker is None:
            blocker = ModificationBlocker(name)
            cls.all_blockers[name] = blocker
            cached_blocker = blocker
        return cached_blocker


def fix_format_string(fmt, dimen=1, key_format='_%s'):
    ''' Fixes a format string so that it contains dimen slots with %s inside
        arguments are:
         --- dimen is the number of slots we need
         --- key_format is the format in which the %s is embedded. By default '_%s'
             for example if each item has to be surrounded by {} set key_format to _{%s}
    '''
    assert (dimen >= 1)
    actual_nb_slots = 0
    curpos = 0
    str_size = len(fmt)
    while curpos < str_size and actual_nb_slots < dimen:
        new_pos = fmt.find('%', curpos)
        if new_pos < 0:
            break
        actual_nb_slots += 1
        if actual_nb_slots >= dimen:
            break
        curpos = new_pos + 2
    # how much slots do we need to add to the end of the string??
    nb_missing = max(0, dimen - actual_nb_slots)
    return fmt + nb_missing * (key_format % '%s')


def is_tuple_w_standard_str(z):
    if isinstance(z, tuple):
        zclass = z.__class__
        return zclass is tuple or not ("__str__" in zclass.__dict__)
    else:
        return False


def str_flatten_tuple(key, sep="_"):
    if is_tuple_w_standard_str(key):
        return sep.join(str(f) for f in key)
    else:
        return str(key)


def compile_naming_function(keys, user_name, dimension=1, key_format=None,
                            _default_key_format='_%s', stringifier=str_flatten_tuple):
    # INTERNAL
    # builds a naming rule from an input , a dimension, and an optional meta-format
    # Makes sure the format string does contain the right number of format slots
    assert user_name is not None

    if is_string(user_name):
        if key_format is None:
            used_key_format = _default_key_format
        elif is_string(key_format):
            # -- make sure some %s is inside, otherwise add it
            if '%s' in key_format:
                used_key_format = key_format
            else:
                used_key_format = key_format + '%s'
        else:  # pragma: no cover
            raise DOcplexException("key format expects string format or None, got: {0!r}".format(key_format))

        fixed_format_string = fix_format_string(user_name, dimension, used_key_format)
        if 1 == dimension:
            return lambda k: fixed_format_string % stringifier(k)
        else:
            # here keys are tuples of size >= 2
            return lambda key_tuple: fixed_format_string % key_tuple

    elif is_function(user_name):
        return user_name

    elif is_iterable(user_name):
        # check that the iterable has same len as keys,
        # otherwise thereis no more default naming and None cannot appear in CPLEX name arrays
        list_names = list(user_name)
        if len(list_names) < len(keys):
            raise DOcplexException("An array of names should have same len as keys, expecting: {0}, go: {1}"
                                   .format(len(keys), len(list_names)))
        key_to_names_dict = {k: nm for k, nm in izip(keys, list_names)}
        # use a closure
        return lambda k: key_to_names_dict[k]  # if k in key_to_names_dict else default_fn()

    else:
        raise DOcplexException('Cannot use this for naming variables: {0!r} - expecting string, function or iterable'
                               .format(user_name))


class _AbstractModelFactory(object):
    def __init__(self, model):
        self._model = model
        self._checker = model._checker
        ordered = model.keep_ordering
        self.term_dict_type = self._term_dict_type_from_ordering(model.keep_ordering)
        self._var_dict_type = dict
        if ordered and not Environment.env_is_python36:
            # FastOrderedDict does not accept zip argument...
            from collections import OrderedDict
            self._var_dict_type = OrderedDict

    def var_dict_type(self, ordered):
        return self._var_dict_type if ordered else dict

    @classmethod
    def _term_dict_type_from_ordering(cls, ordered):
        if not ordered or Environment.env_is_python36:
            return dict
        else:
            from docplex.mp.xcounter import FastOrderedDict
            return FastOrderedDict

    @property
    def ordered(self):
        return self._model._keep_ordering

    def update_ordering(self, ordered):
        self.term_dict_type = self._term_dict_type_from_ordering(ordered)


class ModelFactory(_AbstractModelFactory):
    status_var_fmt = '_bool{{{0:s}}}'

    @staticmethod
    def float_or_default(bound, default_bound):
        return default_bound if bound is None else float(bound)

    def __init__(self, model, engine):
        _AbstractModelFactory.__init__(self, model)

        self._var_container_counter = 0
        self.number_validation_fn = model._checker.get_number_validation_fn()
        self._engine = engine
        self.stringifier = str_flatten_tuple

    @property
    def infinity(self):
        return self._engine.get_infinity()

    def _new_term_dict(self):
        # INTERNAL
        return self.term_dict_type()

    def new_zero_expr(self):
        return ZeroExpr(self._model)

    def fatal(self, msg, *args):
        self._model.fatal(msg, *args)

    def warning(self, msg, *args):
        self._model.warning(msg, args)

    def update_engine(self, engine):
        # the model has already disposed the old engine, if any
        self._engine = engine

    def new_var(self, vartype, lb=None, ub=None, varname=None):
        self_model = self._model
        self._checker.check_var_domain(lb, ub, varname)
        logger = self_model.logger
        rlb = vartype.resolve_lb(lb, logger)
        rub = vartype.resolve_ub(ub, logger)
        used_varname = None if self_model.ignore_names else varname
        var = Var(self_model, vartype, used_varname, rlb, rub, _safe_lb=True, _safe_ub=True)

        idx = self._engine.create_one_variable(vartype, rlb, rub, varname)
        self_model._register_one_var(var, idx, varname)
        return var

    def new_safe_var(self, vartype, lb=None, ub=None, varname=None):
        # INTERNAL
        self_model = self._model
        used_varname = None if self_model.ignore_names else varname
        rlb = vartype._lb if lb is None else lb
        rub = vartype._ub if ub is None else ub
        var = Var(self_model, vartype, used_varname, rlb, rub, _safe_lb=True, _safe_ub=True)
        idx = self._engine.create_one_variable(vartype, rlb, rub, varname)
        self_model._register_one_var(var, idx, used_varname)
        return var

    def new_constraint_status_var(self, ct):
        # INTERNAL
        model = self._model
        binary_vartype = model.binary_vartype
        status_var_fmt = self.status_var_fmt
        if model.ignore_names or status_var_fmt is None:
            varname = None
        else:
            # use name if any else use truncated ct string representation
            base_varname = self.status_var_fmt.format(ct.name or str_maxed(ct, maxlen=20))
            # if name is already taken, use unique index at end to disambiguate
            varname = model._get_non_ambiguous_varname(base_varname)

        svar = Var(model, binary_vartype, lb=0, ub=1, _safe_lb=True, _safe_ub=True, name=varname)
        svar.notify_origin(ct)  # generated

        idx = self._engine.create_one_variable(binary_vartype, 0, 1, varname)
        model._register_one_var(svar, idx, varname)
        return svar

    # --- sequences
    def make_key_seq(self, keys, name):
        # INTERNAL Takes as input a candidate keys input and returns a valid key sequence
        used_name = name
        check_keys = True
        used_keys = []
        if is_iterable(keys):
            if is_pandas_dataframe(keys):
                used_keys = keys.index.values
            elif has_len(keys):
                used_keys = keys
            elif is_iterator(keys):
                used_keys = list(keys)
            else:
                # TODO: make a test for this case.
                self.fatal("Cannot handle iterable var keys: {0!s} : no len() and not an iterator",
                           keys)  # pragma: no cover

        elif is_int(keys) and keys >= 0:
            # if name is str and we have a size, disable automatic names
            used_name = None if name is str else name
            used_keys = range(keys)
            check_keys = False
        else:
            self.fatal("Unexpected var keys: {0!s}, expecting iterable or integer", keys)  # pragma: no cover

        if check_keys and len(used_keys):  # do not check truth value of used_keys: can be a Series!
            self._checker.typecheck_key_seq(used_keys)
        return used_name, used_keys

    def _expand_names(self, keys, user_name, dimension, key_format):
        if user_name is None or self._model.ignore_names:
            # no automatic names, ever
            return []
        else:
            # default_naming_fn = self._model._create_automatic_varname
            actual_naming_fn = compile_naming_function(keys, user_name, dimension, key_format,
                                                       stringifier=self.stringifier)
            computed_names = [actual_naming_fn(key) for key in keys]
            return computed_names

    def _check_bounds(self, nb_vars, bounds, default_bound, true_if_lb):
        nb_bounds = len(bounds)
        bound_name = 'lb' if true_if_lb else 'ub'
        for b, b_value in enumerate(bounds):
            if b_value is not None and not is_number(b_value):
                self.fatal("Variable {2} expect numbers, {0!r} was passed at pos: #{1}",
                           b_value, b, bound_name)
        float_bounds = [self.float_or_default(bv, default_bound) for bv in bounds]

        if nb_bounds > nb_vars:
            self.warning(
                "Variable bounds list too large, required: %d, got: %d." % (nb_vars, nb_bounds))
            return float_bounds[:nb_vars]
        else:
            return float_bounds

    def _expand_bounds(self, keys, var_bound, default_bound, size, true_if_lb):
        ''' Converts raw bounds data (either LB or UB) to CPLEX-compatible bounds list.
            If lbs is None, this is the default, return [].
            If lbs is [] take the default again.
            If it is a number, build a list of size <size> with this number.
            If it is a list, use it if size ok (check numbers??),
            else try it as a function over keys.
        '''
        if var_bound is None:
            # default lb is zero, default ub is infinity
            return []

        elif is_number(var_bound):
            self._checker.typecheck_num(var_bound, caller='in variable bound')
            if true_if_lb:
                if var_bound == default_bound:
                    return []
                else:
                    return [float(var_bound)] * size
            else:
                # ub
                if var_bound >= default_bound:
                    return []
                else:
                    return [float(var_bound)] * size

        elif is_ordered_sequence(var_bound):
            nb_bounds = len(var_bound)
            if nb_bounds < size:
                # see how we can use defaults for those missing bounds
                self.fatal("Variable bounds list is too small, expecting: %d, got: %d" % (size, nb_bounds))
            else:
                return self._check_bounds(size, var_bound, default_bound, true_if_lb)

        elif is_iterator(var_bound):
            # unfold the iterator, as CPLEX needs a list
            return list(var_bound)

        elif isinstance(var_bound, dict):
            dict_bounds = [var_bound.get(k, default_bound) for k in keys]
            return self._check_bounds(size, dict_bounds, default_bound, true_if_lb)
        else:
            # try a function?
            try:
                fn_bounds = [var_bound(k) for k in keys]
                return self._check_bounds(size, fn_bounds, default_bound, true_if_lb)

            except TypeError:
                self._bad_bounds_fatal(var_bound)

            except Exception as e:  # pragma: no cover
                self.fatal("error calling function model bounds: {0!s}, error: {1!s}", var_bound, e)

    def _bad_bounds_fatal(self, bad_bound):
        self.fatal("unexpected variable bound: {0!s}, expecting: None|number|function|iterable", bad_bound)

    def new_multitype_var_list(self, size, vartypes, lbs=None, ubs=None, names=None):
        if not size:
            return []
        mdl = self._model
        assert size == len(vartypes)
        assert size == len(lbs)
        assert size == len(ubs)
        assert size == len(names)

        allvars = [Var(mdl, vartypes[k],
                       names[k] if names[k] else None,
                       lbs[k],
                       ubs[k],
                       _safe_lb=True,
                       _safe_ub=True) for k in fast_range(size)]

        indices = self._engine.create_multitype_variables(size, vartypes, lbs, ubs, names)
        mdl._register_block_vars(allvars, indices, names)
        return allvars

    def var_list(self, keys, vartype, lb, ub, name=None, key_format=None):
        actual_name, fixed_keys = self.make_key_seq(keys, name)
        ctn = self._new_var_container(vartype, key_list=[fixed_keys], lb=lb, ub=ub, name=name)
        return self.new_var_list(ctn, fixed_keys, vartype, lb, ub, actual_name, 1, key_format)

    def new_var_list(self, var_container,
                     key_seq, vartype,
                     lb=None, ub=None,
                     name=str,
                     dimension=1, key_format=None):
        number_of_vars = len(key_seq)
        if 0 == number_of_vars:
            return []

        # compute defaults once
        default_lb = vartype.default_lb
        default_ub = vartype.default_ub

        xlbs = self._expand_bounds(key_seq, lb, default_lb, number_of_vars, true_if_lb=True)
        xubs = self._expand_bounds(key_seq, ub, default_ub, number_of_vars, true_if_lb=False)
        # at this point both list are either [] or have size numberOfVars

        all_names = self._expand_names(key_seq, name, dimension, key_format)

        safe_lbs = not xlbs
        safe_ubs = not xubs
        if xlbs and xubs:
            self._checker.check_vars_domain(xlbs, xubs, all_names)

        mdl = self._model
        allvars = [Var(mdl, vartype,
                       all_names[k] if all_names else None,
                       xlbs[k] if xlbs else default_lb,
                       xubs[k] if xubs else default_ub,
                       container=var_container,
                       _safe_lb=safe_lbs,
                       _safe_ub=safe_ubs) for k in fast_range(number_of_vars)]

        # query the engine for a list of indices.
        indices = self._engine.create_variables(len(key_seq), vartype, xlbs, xubs, all_names)
        mdl._register_block_vars(allvars, indices, all_names)
        return allvars

    def _make_var_dict(self, keys, var_list, ordered):
        self._checker.check_for_duplicate_keys(keys)
        _dict_type = self.var_dict_type(ordered)
        vdict = _dict_type(izip(keys, var_list))
        return vdict

    def new_var_dict(self, keys, vartype, lb, ub, name, key_format, ordered=False):
        actual_name, key_seq = self.make_key_seq(keys, name)

        ctn = self._new_var_container(vartype, key_list=[key_seq], lb=lb, ub=ub, name=name)
        var_list = self.new_var_list(ctn, key_seq, vartype, lb, ub, actual_name, 1, key_format)
        return self._make_var_dict(key_seq, var_list, ordered)

    def new_var_multidict(self, seq_of_key_seqs, vartype, lb, ub, name, key_format=None, ordered=False):
        # ---
        fixed_keys = [self.make_key_seq(ks, name)[1] for ks in seq_of_key_seqs]
        # the sequence of keysets should answer to len(no generators here)
        dimension = len(fixed_keys)
        if dimension < 1:
            self.fatal("len of key sequence must be >= 1, got: {0}", dimension)  # pragma: no cover

        # create cartesian product of keys...
        all_key_tuples = list(product(*fixed_keys))
        # check empty list
        if not all_key_tuples:
            self.fatal('multidict has no keys to index the variables')

        # always pass a sequence of sequences
        ctn = self._new_var_container(vartype, key_list=fixed_keys, lb=lb, ub=ub, name=name)
        cube_vars = self.new_var_list(ctn, all_key_tuples, vartype, lb, ub, name, dimension, key_format)

        return self._make_var_dict(keys=all_key_tuples, var_list=cube_vars, ordered=ordered)

    def new_var_df(self, keys1, keys2, vartype, lb=None, ub=None, name=None):  # pragma: no cover
        try:
            from pandas import DataFrame
        except ImportError:
            DataFrame = None
            self.fatal('make_var_df() requires pandas module - not found')

        _, row_keys = self.make_key_seq(keys1, name)
        _, col_keys = self.make_key_seq(keys2, name)
        matrix_keys = [(k1, k2) for k1 in row_keys for k2 in col_keys]
        ctn = self._new_var_container(vartype, key_list=matrix_keys, lb=lb, ub=ub, name=name)
        lvars = self.new_var_list(ctn, matrix_keys, vartype, lb, ub, name, dimension=2, key_format=None)
        # TODO: see how to do without this temp dict
        dd = dict(izip(matrix_keys, lvars))
        # row-oriented dict
        rowd = {row_k: [dd[row_k, col_k] for col_k in col_keys] for row_k in row_keys}
        vdtf = DataFrame.from_dict(rowd, orient='index', dtype='object')
        # convert to string or not?
        vdtf.columns = col_keys
        return vdtf

    def _new_constant_expr(self, cst, safe_number=True):
        k = cst
        if not safe_number:
            self_number_validation_fn = self.number_validation_fn
            if self_number_validation_fn:
                k = self_number_validation_fn(cst)
        return ConstantExpr(self._model, k)

    use_constant_expr = True

    def constant_expr(self, cst, safe_number=False):
        if not cst:
            return self.new_zero_expr()
        else:
            return self._new_constant_expr(cst, safe_number=safe_number)
        # else:  # pragma: no cover
        #     if safe_number:
        #         k = cst
        #     else:
        #         self_number_validation_fn = self.number_validation_fn
        #         k = self_number_validation_fn(cst) if self_number_validation_fn else cst
        #     return LinearExpr(self._model, e=None, constant=k, safe=True)

    def linear_expr(self, arg=None, constant=0, name=None, safe=False):
        return LinearExpr(self._model, arg, constant, name, safe=safe)

    # def to_valid_number(self, e, checked_num=False, context_msg=None, infinity=1e+20):
    #     return self._checker.to_valid_number(e, checked_num, context_msg, infinity)

    _operand_types = (AbstractLinearExpr, Var, ZeroExpr)

    @staticmethod
    def _is_operand(arg, accept_numbers=False):
        return isinstance(arg, Operand) or (accept_numbers and is_number(arg))

    def _to_linear_operand(self, e, force_clone=False, msg=None):
        if isinstance(e, LinearOperand):
            if force_clone:
                return e.clone()
            else:
                return e
        elif is_number(e):
            return self.constant_expr(cst=e, safe_number=False)
        else:
            try:
                return e.to_linear_expr()
            except AttributeError:
                # delegate to the factory
                return self.linear_expr(e)
            except DocplexQuadToLinearException as qe:
                used_msg = msg.format(e) if msg else qe.message
                raise DOcplexException(used_msg)

    def _to_linear_expr(self, e, linexpr_class=LinearExpr, force_clone=False):
        # TODO: replace by to_linear_operand
        if isinstance(e, linexpr_class):
            if force_clone:
                return e.clone()
            else:
                return e
        elif isinstance(e, self._operand_types):
            return e.to_linear_expr()
        elif is_number(e):
            return self.constant_expr(cst=e, safe_number=False)
        else:
            try:
                return e.to_linear_expr()
            except AttributeError:
                # delegate to the factory
                return self.linear_expr(e)

    def _to_expr(self, e):
        # INTERNAL
        if hasattr(e, "iter_terms"):
            return e
        elif is_number(e):
            return self.constant_expr(cst=e, safe_number=True)
        else:
            try:
                return e.to_expr()
            except AttributeError:
                self.fatal("cannot convert to expression: {0!r}", e)

    def new_binary_constraint(self, lhs, sense, rhs, name=None):
        ctsense = ComparisonType.parse(sense)
        return self._new_binary_constraint(lhs, ctsense, rhs, name)

    def _new_binary_constraint(self, lhs, sense, rhs, name=None):
        # noinspection PyPep8
        left_expr = self._to_linear_operand(lhs, msg="LinearConstraint. expects linear expressions, {0} was passed")
        right_expr = self._to_linear_operand(rhs, msg="LinearConstraint. expects linear expressions, {0} was passed")
        self._checker.typecheck_two_in_model(self._model, left_expr, right_expr, "new_binary_constraint")
        ct = LinearConstraint(self._model, left_expr, sense, right_expr, name)
        return ct

    def new_le_constraint(self, e, rhs, ctname=None):
        return self._new_binary_constraint(e, ComparisonType.LE, rhs, name=ctname)

    def new_eq_constraint(self, e, rhs, ctname=None):
        return self._new_binary_constraint(e, ComparisonType.EQ, rhs, name=ctname)

    def new_ge_constraint(self, e, rhs, ctname=None):
        return self._new_binary_constraint(e, ComparisonType.GE, rhs, name=ctname)

    def new_neq_constraint(self, lhs, rhs, ctname=None):
        m = self._model
        left_expr = self._to_linear_operand(lhs,
                                            msg="The `!=` operator requires two linear expressions, {0} was passed (left)")
        right_expr = self._to_linear_operand(rhs,
                                             msg="The `!=` operator requires two linear expressions, {0} was passed (right)")
        StaticTypeChecker.typecheck_discrete_expression(m, msg="NotEqualConstraint", expr=left_expr)
        StaticTypeChecker.typecheck_discrete_expression(m, msg="NotEqualConstraint", expr=right_expr)
        self._checker.typecheck_two_in_model(m, left_expr, right_expr, "new_binary_constraint")
        negated_ct = self.new_eq_constraint(lhs, rhs)
        ct = NotEqualConstraint(self._model, negated_ct, ctname)
        return ct

    def _check_range_feasibility(self, lb, ub, expr):
        # INTERNAL
        if not lb <= ub:
            self._model.error("infeasible range constraint, lb={0}, ub={1}, expr={2}", lb, ub, expr)

    def new_range_constraint(self, lb, expr, ub, name=None, check_feasible=True):
        self._check_range_feasibility(lb, ub, expr)
        linexpr = self._to_linear_operand(expr)
        rng = RangeConstraint(self._model, linexpr, lb, ub, name)
        linexpr.notify_used(rng)
        return rng

    def new_indicator_constraint(self, binary_var, linear_ct, true_value=1, name=None):
        # INTERNAL
        indicator_ct = IndicatorConstraint(self._model, binary_var, linear_ct, true_value, name)
        return indicator_ct

    def new_equivalence_constraint(self, binary_var, linear_ct, true_value=1, name=None):
        # INTERNAL
        equiv_ct = EquivalenceConstraint(self._model, binary_var, linear_ct, true_value, name)
        return equiv_ct

    def new_if_then_constraint(self, if_ct, then_ct, negate=False):
        # INTERNAL
        indicator_ct = IfThenConstraint(self._model, if_ct, then_ct, negate=negate)
        return indicator_ct

    def new_batch_equivalence_constraints(self, bvars, linear_cts, active_values, names):
        return [self.new_equivalence_constraint(bv, lct, active, name)
                for bv, lct, active, name in izip(bvars, linear_cts, active_values, names)]

    def new_batch_indicator_constraints(self, bvars, linear_cts, active_values, names):
        return [self.new_indicator_constraint(bv, lct, active, name)
                for bv, lct, active, name in izip(bvars, linear_cts, active_values, names)]

    def new_binary_constraint_or(self, ct1, arg2):
        orexpr = self.new_logical_or_expr([ct1, arg2])
        return orexpr

    def new_binary_constraint_and(self, ct, other):
        and_expr = self.new_logical_and_expr([ct, other])
        return and_expr

    # updates

    def update_linear_constraint_exprs(self, ct, expr_event):
        ct_event = UpdateEvent.LinearConstraintRhs if expr_event is UpdateEvent.ExprConstant \
            else UpdateEvent.LinearConstraintGlobal
        self._engine.update_constraint(ct, ct_event)

    def update_indicator_constraint_expr(self, ind, expr, event):
        self._engine.update_constraint(ind, UpdateEvent.IndicatorLinearConstraint, expr)

    def check_expr_discrete_lock(self, expr, arg):
        # INTERNAL
        if expr.is_discrete_locked():
            self.fatal('Expression: {0} is used in equivalence, cannot be modified', expr)

    def set_linear_constraint_expr_from_pos(self, lct, pos, new_expr, update_subscribers=True):
        # INTERNAL
        # pos is 0 for left, 1 for right
        new_operand = self._to_linear_operand(e=new_expr, force_clone=False)
        lct._check_editable(new_operand, self._engine)

        old_expr = lct.get_expr_from_pos(pos)
        exprs = [lct._left_expr, lct._right_expr]
        exprs[pos] = new_operand
        # -- event
        if old_expr.is_constant() and new_operand.is_constant():
            event = UpdateEvent.LinearConstraintRhs
        else:
            event = UpdateEvent.LinearConstraintGlobal
        # ---
        self._engine.update_constraint(lct, event, *exprs)
        lct.set_expr_from_pos(pos, new_operand)
        if update_subscribers:
            # -- update  subscribers
            old_expr.notify_unsubscribed(lct)
            new_operand.notify_used(lct)

    def set_linear_constraint_right_expr(self, ct, new_rexpr):
        self.set_linear_constraint_expr_from_pos(ct, pos=1, new_expr=new_rexpr)

    def set_linear_constraint_left_expr(self, ct, new_lexpr):
        self.set_linear_constraint_expr_from_pos(ct, pos=0, new_expr=new_lexpr)

    def set_linear_constraint_sense(self, ct, arg_newsense):
        new_sense = ComparisonType.parse(arg_newsense)
        if new_sense != ct.sense:
            self._engine.update_constraint(ct, UpdateEvent.ConstraintSense, new_sense)
            ct._internal_set_sense(new_sense)

    def set_range_constraint_lb(self, rngct, new_lb):
        self.set_range_constraint_bounds(rngct, new_lb, None)

    def set_range_constraint_ub(self, rngct, new_ub):
        self.set_range_constraint_bounds(rngct, None, new_ub)

    def set_range_constraint_bounds(self, rngct, new_lb, new_ub):
        lb_to_use = rngct.lb if new_lb is None else new_lb
        ub_to_use = rngct.ub if new_ub is None else new_ub
        # assuming the new bound has been typechecked already..
        self._engine.update_range_constraint(rngct, UpdateEvent.RangeConstraintBounds, lb_to_use, ub_to_use)
        if new_lb is not None:
            rngct._internal_set_lb(new_lb)
        if new_ub is not None:
            rngct._internal_set_ub(new_ub)
        self._check_range_feasibility(rngct.lb, rngct.ub, rngct.expr)

    def set_range_constraint_expr(self, rngct, new_expr):
        new_op = self._to_linear_operand(new_expr)
        self._engine.update_range_constraint(rngct, UpdateEvent.RangeConstraintExpr, new_op)
        old_expr = rngct.expr
        rngct._expr = new_expr
        old_expr.notify_unsubscribed(rngct)

    # ---------------------
    def _new_logical_expr(self, args, ctor, empty_value):
        bvars = [arg.as_logical_operand() for arg in args]
        # assume bvars is a sequence of binary vars
        nb_args = len(bvars)
        if not nb_args:
            return self._new_constant_expr(cst=empty_value, safe_number=True)
        elif 1 == nb_args:
            return bvars[0]
        else:
            return ctor(self._model, bvars)

    def new_logical_and_expr(self, args):
        return self._new_logical_expr(args, ctor=LogicalAndExpr, empty_value=1)

    def new_logical_or_expr(self, args):
        return self._new_logical_expr(args, ctor=LogicalOrExpr, empty_value=0)

    def new_logical_not_expr(self, arg):
        return LogicalNotExpr(self._model, arg)

    def logical_expr_to_constraint(self, logical_expr, name=None):
        # generate a constraint expr == 1
        ct1 = self._new_binary_constraint(lhs=logical_expr, rhs=1.0, sense=ComparisonType.EQ, name=name)
        return ct1

    def _new_min_max_expr(self, expr_class, builtin_fn, empty_value, *args):
        nb_args = len(args)
        if 0 == nb_args:
            return empty_value
        elif 1 == nb_args:
            return args[0]
        elif all(is_number(a) for a in args):
            # if all args are numbers, simply compute a number
            return builtin_fn(args)  # a number
        else:
            return expr_class(self._model, [self._to_linear_operand(a) for a in args])

    def new_max_expr(self, *args):
        return self._new_min_max_expr(MaximumExpr, max, -self.infinity, *args)

    def new_min_expr(self, *args):
        return self._new_min_max_expr(MinimumExpr, min, self.infinity, *args)

    def new_abs_expr(self, e):
        if is_number(e):
            return abs(e)
        else:
            self_model = self._model
            return AbsExpr(self_model, self._to_linear_operand(e))

    def resync_whole_model(self):
        self_model = self._model
        self_engine = self._engine

        for var in self_model.iter_variables():
            # do not call create_one_var public API
            # or resync would loop
            idx = self_engine.create_one_variable(var.vartype, var.lb, var.ub, var.name)
            if idx != var.get_index():  # pragma: no cover
                print("index discrepancy: {0!s}, new index= {1}, old index={2}"
                      .format(var, idx, var.get_index()))

        for ct in self_model.iter_constraints():
            if isinstance(ct, LinearConstraint):
                self_engine.create_linear_constraint(ct)
            elif isinstance(ct, RangeConstraint):
                self_engine.create_range_constraint(ct)
            elif isinstance(ct, IndicatorConstraint):
                self_engine.create_logical_constraint(ct, is_equivalence=False)
            elif isinstance(ct, EquivalenceConstraint):
                self_engine.create_logical_constraint(ct, is_equivalence=True)
            elif isinstance(ct, QuadraticConstraint):
                self_engine.create_quadratic_constraint(ct)
            else:
                self_model.fatal("Unexpected constraint type: {0!s} - ignored", type(ct))  # pragma: no cover

        # send objective
        self_engine.set_objective_sense(self_model.objective_sense)
        if self_model.has_multi_objective():
            multi_objective = self_model._multi_objective
            exprs = multi_objective.exprs
            nb_exprs = len(exprs)
            self_engine.set_multi_objective_exprs(new_multiobjexprs=exprs,
                                                  old_multiobjexprs=[],
                                                  priorities=multi_objective.priorities,
                                                  weights=multi_objective.weights,
                                                  abstols=MultiObjective.as_optional_sequence(multi_objective.abstols,
                                                                                              nb_exprs),
                                                  reltols=MultiObjective.as_optional_sequence(multi_objective.reltols,
                                                                                              nb_exprs),
                                                  objnames=multi_objective.names)
        else:
            self_engine.set_objective_expr(self_model.objective_expr, old_objexpr=None)

    def new_sos(self, dvars, sos_type, weights, name):
        # INTERNAL
        new_sos = SOSVariableSet(model=self._model, variable_sequence=dvars, sos_type=sos_type, weights=weights,
                                 name=name)
        return new_sos

    def new_piecewise(self, pwl_def, name):
        self_model = self._model
        # Note that this object is specific only to DOcplex. It does not map to a Cplex object.
        pwl = PwlFunction(self_model, pwl_def=pwl_def, name=name)
        return pwl

    def new_pwl_expr(self, pwl_func, e, usage_counter, y_var=None, add_counter_suffix=True, resolve=True):
        return PwlExpr(self._model, pwl_func, e,
                       usage_counter,
                       add_counter_suffix=add_counter_suffix,
                       y_var=y_var,
                       resolve=resolve)

    def new_pwl_constraint(self, pwl_expr, name=None):
        self_model = self._model
        return PwlConstraint(self_model, pwl_expr, name)

    @staticmethod
    def default_objective_sense():
        return ObjectiveSense.Minimize

    def new_kpi(self, kpi_arg, name_arg):
        # make a name
        if name_arg:
            publish_name = name_arg
        elif hasattr(kpi_arg, 'name') and kpi_arg.name:
            publish_name = kpi_arg.name
        else:
            publish_name = str_maxed(kpi_arg, maxlen=32)
        new_kpi = KPI.new_kpi(self._model, kpi_arg, publish_name)
        return new_kpi

    def _new_constraint_block2(self, cts, ctnames):
        posted_cts = []
        prepfn = self._model._prepare_constraint
        checker = self._checker
        check_trivials = checker.check_trivial_constraints()

        if is_string(ctnames):
            from docplex.mp.utils import _AutomaticSymbolGenerator
            # no separator added, use a terminal "_" if need be
            ctnames = _AutomaticSymbolGenerator(ctnames)

        for ct, ctname in izip_longest(cts, ctnames):  # use izip_longest so as not to forget any ct
            if ct is None:  # izip stops
                break

            checker.typecheck_linear_constraint(ct, accept_range=False)
            checker.typecheck_string(ctname, accept_none=True, accept_empty=False, caller="Model.add_constraints()")
            if prepfn(ct, ctname, check_for_trivial_ct=check_trivials):
                posted_cts.append(ct)
        self._post_constraint_block(posted_cts)
        return posted_cts

    def _new_constraint_block1(self, cts):
        posted_cts = []
        checker = self._checker
        filterfn = self._model._prepare_constraint
        check_trivial = self._checker.check_trivial_constraints()
        # look first
        ctseq = list(cts)
        if not ctseq:
            return []

        try:
            # noinspection PyUnusedLocal
            ct, ctname = ctseq[0]
            tuple_mode = True
        except (TypeError, ValueError):
            # TypeError is for non-tuple
            # ValueError is for nonlength-2 tuples
            # not a tuple: we have only constraints and no names
            tuple_mode = False

        if tuple_mode:
            ctseq2 = self._checker.typecheck_linear_constraint_name_tuple_seq(ctseq)
            for ct, ctname in ctseq2:
                if filterfn(ct, ctname, check_for_trivial_ct=check_trivial, arg_checker=checker):
                    posted_cts.append(ct)
        else:
            checker.typecheck_constraint_seq(ctseq, check_linear=True, accept_range=True)
            for ct in ctseq:
                if filterfn(ct, ctname=None, check_for_trivial_ct=check_trivial, arg_checker=checker):
                    posted_cts.append(ct)
        self._post_constraint_block(posted_cts)
        return posted_cts

    def _post_constraint_block(self, posted_cts):
        if posted_cts:
            ct_indices = self._engine.create_block_linear_constraints(posted_cts)
            self._model._register_block_cts(self._model._linct_scope, posted_cts, ct_indices)

    # --- range block

    def new_range_block(self, lbs, exprs, ubs, names):
        try:
            n_exprs = len(exprs)
            if n_exprs != len(lbs):  # pragma: no cover
                self.fatal('incorrect number of expressions: expecting {0}, got: {1}'.format(len(lbs), n_exprs))

        except TypeError:
            pass  # no len available.
        if not names:
            names = generate_constant(None, None)

        ranges = [self.new_range_constraint(lb, exp, ub, name) for lb, exp, ub, name in
                      izip(lbs, exprs, ubs, names)]
        # else:
        #     ranges = [self.new_range_constraint(lb, exp, ub) for lb, exp, ub in izip(lbs, exprs, ubs)]
        self._post_constraint_block(ranges)
        return ranges

    def new_solution(self, var_value_dict=None, name=None, objective_value=None, **kwargs):
        keep_zeros = kwargs.get('keep_zeros', True)

        return SolveSolution(model=self._model, obj=objective_value,
                             var_value_map=var_value_dict, name=name,
                             keep_zeros=keep_zeros)

    def _new_var_container(self, vartype, key_list, lb, ub, name):
        # INTERNAL
        ctn = _VariableContainer(vartype, key_list, lb, ub, name)
        old_varctn_counter = self._var_container_counter
        ctn._index = old_varctn_counter
        ctn._index_offset = self._model.number_of_variables  # nb of variables before ctn
        self._var_container_counter = old_varctn_counter + 1

        self._model._add_var_container(ctn)
        return ctn


class _VariableContainer(object):
    def __init__(self, vartype, keys_seq, lb, ub, name):
        self._index = 0
        self._index_offset = 0
        self._vartype = vartype
        self._keyss = keys_seq
        self._lb = lb
        self._ub = ub
        self._name = name
        self._name_str = None

    @property
    def index(self):
        return self._index

    def copy(self, target_model):
        copied_ctn = self.__class__(self.vartype, self._keyss, self.lb, self.ub, self._name)
        return copied_ctn

    def copy_relaxed(self, target_model):
        copied_ctn = self.__class__(target_model.continuous_vartype, self._keyss, self.lb, self.ub, self._name)
        return copied_ctn

    def keys(self):
        return self._keyss

    @property
    def vartype(self):
        return self._vartype

    @property
    def nb_dimensions(self):
        return len(self._keyss)

    @property
    def namer(self):
        return self._name

    @property
    def lb(self):
        return self._lb

    @property
    def ub(self):
        return self._ub

    @property
    def name(self):
        """
        Try to extract a name string from the initial container name.
        handles strings with or without formats, arrays, function.

        :return: A string.
        """
        return self._lazy_compute_name_string()

    def iter_keys(self):
        if 1 == self.nb_dimensions:
            return iter(self._keyss[0])
        else:
            return product(*self._keyss)

    def _lazy_compute_name_string(self):
        if self._name_str is not None:
            return self._name_str
        else:
            raw_name = self._name
            if is_string(raw_name):
                # drop opl-style formats
                s_name = raw_name.replace("({%s})", "")
                # purge fields
                pos_pct = raw_name.find('%')
                if pos_pct >= 0:
                    s_name = raw_name[:pos_pct - 1]
                elif raw_name.find('{') > 0:
                    pos = raw_name.find('{')
                    s_name = raw_name[:pos - 1]
            elif is_iterable(raw_name):
                from os.path import commonprefix
                s_name = commonprefix(raw_name)
            else:
                # try a function
                from os.path import commonprefix
                namefn = raw_name
                try:
                    all_names = [namefn(k) for k in self.iter_keys()]
                    s_name = commonprefix(all_names)

                except TypeError:
                    s_name = ''

            self._name_str = s_name
            return s_name

    def get_var_key(self, dvar):
        # INTERNAL
        # containers store expanded keys (as tuples).
        dvar_index = dvar.get_index()
        relative_offset = dvar_index - self._index_offset
        if self.nb_dimensions == 1:
            try:
                return self._keyss[0][relative_offset]
            except IndexError:  # pragma: no cover
                return None
        else:
            return next(islice(product(*self._keyss), relative_offset, None), None)

    def shape(self):
        return tuple(len(k) for k in self._keyss)

    @property
    def dimension_string(self):
        dim_string = "".join(["[%d]" % s for s in self.shape()])
        return dim_string

    def to_string(self):
        # dvar xxx
        dim_string = self.dimension_string
        ctname = self._name or 'x'
        return "dvar {0} {1} {2}".format(self.vartype.short_name, ctname, dim_string)

    def __str__(self):
        return self.to_string()



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

import sys
from io import StringIO

from enum import Enum

from docplex.mp.operand import Operand
from docplex.mp.utils import is_number, is_string, str_maxed
from docplex.mp.format import LP_format


class ModelingObjectBase(object):
    """ModelingObjectBase()

    Parent class for all modeling objects (variables and constraints).

    This class is not intended to be instantiated directly.
    """

    __array_priority__ = 100

    __slots__ = ("_name", "_model")

    def __init__(self, model, name=None):
        self._name = name
        self._model = model

    @property
    def name(self):
        """ This property is used to get or set the name of the modeling object.

        """
        return self._name

    @name.setter
    def name(self, new_name):
        self.set_name(new_name)

    def get_name(self):
        return self._name

    def set_name(self, name):
        self.check_name(name)
        self._set_name(name)

    def _set_name(self, name):
        self._name = name

    @property
    def safe_name(self):
        return self._name or ''

    def check_name(self, new_name):
        # INTERNAL: basic method for checking names.
        pass  # pragma: no cover

    def check_lp_name(self, new_name):
        # ModelingObject.check_name(self, new_name)
        if not is_string(new_name) or not new_name:
            self.fatal("Variable name accepts only non-empty strings, {0!r} was passed", new_name)
        elif new_name.find(' ') >= 0:
            self.warning("Variable name contains blank space, var: {0!s}, name: \'{1!s}\'", self, new_name)
        elif not LP_format.is_lp_compliant(new_name):
            self.warning(
                "Candidate variable name is not LP-compliant: '{1}', old_name was: {0} (name will be changed to x<nn>)",
                self.name, new_name)

    def has_name(self):
        """ Checks whether the object has a name.

        Returns:
            True if the object has a name.

        """
        return self._name is not None

    def has_user_name(self):
        """ Checks whether the object has a valid name given by the user.

        Returns:
            True if the object has a valid name given by the user.

        """
        return self.has_name()

    @property
    def model(self):
        """
        This property returns the :class:`docplex.mp.model.Model` to which the object belongs.
        """
        return self._model

    def _get_model(self):
        return self._model

    def get_linear_factory(self):
        return self._model._lfactory

    def get_quadratic_factory(self):
        return self._model._qfactory

    def is_in_model(self, model):
        return model and self._model is model

    def _check_model_has_solution(self):
        self.model.check_has_solution()

    def fatal(self, msg, *args):
        self.error_handler.fatal(msg, args)

    def error(self, msg, *args):
        self.error_handler.error(msg, args)

    def warning(self, msg, *args):
        self.error_handler.warning(msg, args)

    def trace(self, msg, *args):
        self.error_handler.trace(msg, args)

    @property
    def error_handler(self):
        return self._model.error_handler

    def truncated_str(self):
        return str_maxed(self, maxlen=self._model._max_repr_len)

    def zero_expr(self):
        # INTERNAL
        return self._model._lfactory.new_zero_expr()

    def _unsupported_binary_operation(self, lhs, op, rhs):
        self.fatal("Unsupported operation: {0!s} {1:s} {2!s}", lhs, op, rhs)

    def is_quad_expr(self):
        return False

    def __unicode__(self):
        return self.to_string()

    def __str__(self):
        if sys.version_info[0] == 2:
            return self.__unicode__().encode('utf-8')
        else:  # pragma: no cover
            return self.__unicode__()


class ModelingObject(ModelingObjectBase):
    __slots__ = ("_index", "_origin", "_container")

    @staticmethod
    def is_valid_index(idx):
        # INTERNAL: This is where the valid index check is performed
        return idx >= 0

    _invalid_index = -2

    # noinspection PyMissingConstructor
    def __init__(self, model, name=None, index=_invalid_index):
        #  ModelingObjectBase.__init__(self, model, name)
        self._model = model
        self._name = name
        self._index = index

    def is_generated(self):
        """ Checks whether this object has been generated by another modeling object.

        If so, the origin object is stored in the ``_origin`` attribute.

        Returns:
            True if the objects has been generated.
        """
        return hasattr(self, '_origin') and self._origin is not None

    def notify_origin(self, origin):
        if origin is not None:
            self._origin = origin

    def origin(self):
        return getattr(self, '_origin', None)

    def __hash__(self):
        return id(self)

    @property
    def unchecked_index(self):
        return self._index

    def get_index(self):
        return self._index

    def set_index(self, idx):
        self._index = idx

    def has_valid_index(self):
        return self._index >= 0

    def _set_invalid_index(self):
        self._index = self._invalid_index

    @property
    def safe_index(self):
        if not self.has_valid_index():
            self.fatal("Modeling object {0!s} has invalid index: {1:d}", self, self._index)  # pragma: no cover
        return self._index

    index = property(get_index, set_index)

    def get_container(self):
        # INTERNAL
        return getattr(self, '_container', None)


class Expr(ModelingObjectBase, Operand):
    """Expr()

    Parent class for all expression classes.
    """
    __slots__ = ()

    def __init__(self, model, name=None):
        ModelingObjectBase.__init__(self, model, name)

    def clone(self):
        raise NotImplementedError  # pragma: no cover

    def copy(self, target_model, var_mapping):
        # internal
        raise NotImplementedError  # pragma: no cover

    def number_of_variables(self):
        """
        Returns:
            integer: The number of variables in the expression.
        """
        return sum(1 for _ in self.iter_variables())  # pragma: no cover

    def contains_var(self, dvar):
        """ Checks whether a variable is present in the expression.

        :param: dvar (:class:`docplex.mp.linear.Var`): A decision variable.

        Returns:
            Boolean: True if the variable is present in the expression, else False.
        """
        for v in self.iter_variables():
            if dvar is v:
                return True
        else:
            return False

    def to_string(self, nb_digits=None, use_space=False):
        oss = StringIO()
        if nb_digits is None:
            nb_digits = self.model.float_precision
        self.to_stringio(oss, nb_digits=nb_digits, use_space=use_space)
        return oss.getvalue()

    def to_stringio(self, oss, nb_digits, use_space, var_namer=lambda v: v.name):
        raise NotImplementedError  # pragma: no cover

    def is_model_ordered(self):
        return self._model._keep_ordering


    def _num_to_stringio(self, oss, num, ndigits=None, print_sign=False, force_plus=False, use_space=False):
        k = num
        if print_sign:
            if k < 0:
                sign = u'-'
                k = -k
            elif k > 0 and force_plus:
                # force a plus
                sign = u'+'
            else:
                sign = None
            if use_space:
                oss.write(u' ')
            if sign:
                oss.write(sign)
            if use_space:
                oss.write(u' ')
        # INTERNAL
        ndigits = ndigits or self.model.float_precision
        if k == int(k):
            oss.write(u'%d' % k)
        else:
            # use second arg as nb digits:
            oss.write(u"{0:.{1}f}".format(k, ndigits))

    # def __pos__(self):
    #     # + e is identical to e
    #     return self

    def is_discrete(self):
        raise NotImplementedError  # pragma: no cover

    def is_zero(self):
        return False

    constant = property(Operand.get_constant)

    @property
    def float_precision(self):
        return 0 if self.is_discrete() else self.model.float_precision

    def _round_if_discrete(self, raw_value):
        return self.model._round_element_value_if_necessary(self, raw_value)

    def _get_solution_value(self, s=None):
        # INTERNAL: compute solution value.
        raise NotImplementedError  # pragma: no cover

    @property
    def solution_value(self):
        self._check_model_has_solution()
        return self._get_solution_value()

    def __pow__(self, power):
        # INTERNAL
        # power must be checke in {0, 1, 2}
        self.model.typecheck_as_power(self, power)
        if 0 == power:
            return 1
        elif 1 == power:
            return self
        else:
            return self.square()

    def square(self):
        # redefine for each class of expression
        return None  # pragma: no cover

    def __gt__(self, e):
        """ The strict > operator is not supported
        """
        self.model.unsupported_relational_operator_error(self, ">", e)

    def __lt__(self, e):
        """ The strict < operator is not supported
        """
        self.model.unsupported_relational_operator_error(self, "<", e)


# --- Priority class used for relaxation
class Priority(Enum):
    """
    This enumerated class defines the priorities: VERY_LOW, LOW, MEDIUM, HIGH, VERY_HIGH, MANDATORY.
    """

    def __new__(cls, value, print_name):
        obj = object.__new__(cls)
        # predefined
        obj._value_ = value
        obj._print_name = print_name
        return obj

    VERY_LOW = 100, 'Very Low'
    LOW = 200, 'Low'
    MEDIUM = 300, 'Medium'
    HIGH = 400, 'High'
    VERY_HIGH = 500, 'Very High'
    MANDATORY = 999999999, 'Mandatory'

    def __repr__(self):
        return 'Priority<{}>'.format(self.name)

    @property
    def cplex_preference(self):
        return self._get_geometric_preference_factor(base=10.0)

    def _get_geometric_preference_factor(self, base):
        # INTERNAL: returns a CPLEX preference factor as a power of "base"
        # MEDIUM priority is the balance point with a preference of 1.
        assert is_number(base)

        if self.is_mandatory():
            return 1e+20
        else:
            # noinspection PyTypeChecker
            medium_index = Priority.MEDIUM.value / 100
            # pylint complains about no value member but is wrong!
            diff = self.value / 100 - medium_index
            factor = 1.0
            pdiff = diff if diff >= 0 else -diff
            for _ in range(0, int(pdiff)):
                factor *= base
            return factor if diff >= 0 else 1.0 / factor

    def less_than(self, other):
        assert isinstance(other, Priority)
        return self.value < other.value

    def __lt__(self, other):
        return self.less_than(other)

    def __gt__(self, other):
        return other.less_than(self)

    def is_mandatory(self):
        return self == Priority.MANDATORY

    @classmethod
    def parse(cls, arg, logger, accept_none=True, do_raise=True):
        ''' Converts its argument to a ``Priority`` object.

        Returns `default_priority` if the text is not a string, empty, or does not match.

        Args;
            arg: The argument to convert.

            logger: An error logger

            accept_none: True if None is a possible value. Typically,
                Constraint.set_priority accepts None as a way to
                remove the constraint's own priority.

            do_raise: A Boolean flag indicating if an exception is to be raised if the value
                is not recognized.

        Returns:
            A Priority enumerated object.
        '''
        if isinstance(arg, cls):
            return arg
        elif is_string(arg):
            key = arg.lower()
            # noinspection PyTypeChecker
            for p in cls:
                if key == p.name.lower() or key == str(p.value):
                    return p
            else:
                if do_raise:
                    logger.fatal('String does not match priority type: {}', arg)
                else:
                    logger.error('String does not match priority type: {}', arg)
                    return None
                return None
        elif accept_none and arg is None:
            return None
        else:
            logger.fatal('Cannot convert to a priority: {0!s}'.format(arg))


# ---


class UserPriority(object):

    def __init__(self, pref, name=None):
        assert pref >= 0
        self._preference = pref
        self._name = name

    @property
    def cplex_preference(self):
        return self._preference

    # noinspection PyMethodMayBeStatic
    def is_mandatory(self):
        return False

    @property
    def name(self):
        return self._name or '_user_'

    @property
    def value(self):
        return self._preference

    def __str__(self):
        name = self._name
        sname = '%s: ' % name if name else ''
        return 'UserPriority({0}{1})'.format(sname, self._preference)


# noinspection PyUnusedLocal,PyPropertyAccess

class _SubscriptionMixin(object):
    __slots__ = ()

    # INTERNAL:
    # This class is absolutely not meant to be directly instantiated
    # but used as a mixin

    @classmethod
    def _new_empty_subscribers(cls):
        return []

    def notify_used(self, user):
        # INTERNAL
        self._subscribers.append(user)

    notify_subscribed = notify_used

    def notify_unsubscribed(self, subscriber):
        # 1 find index
        for s, sc in enumerate(self._subscribers):
            if sc is subscriber:
                del self._subscribers[s]
                break

    def clear_subscribers(self):
        self._subscribers = []

    def is_in_use(self):
        return bool(self._subscribers)

    @property
    def nb_subscribers(self):
        return len(self._subscribers)

    def is_shared(self):
        return self.nb_subscribers >= 2

    def is_used_by(self, obj):
        # lists are not optimal here, but we favor insertion: append is faster than set.add
        return any(obj is sc for sc in self.iter_subscribers())

    def notify_modified(self, event):
        for s in self._subscribers:
            s.notify_expr_modified(self, event)

    def iter_subscribers(self):
        return iter(self._subscribers)

    def notify_replaced(self, new_expr):
        for s in self._subscribers:
            s.notify_expr_replaced(self, new_expr)

    def grab_subscribers(self, other):
        # grab subscribers from another expression
        # typically when an expression is replaced by another.
        for s in other.iter_subscribers():
            self._subscribers.append(s)
        # delete all subscriptions on old
        other.clear_subscribers()


class _BendersAnnotatedMixin(object):
    __slots__ = ()

    def set_benders_annotation(self, group):
        self._model.set_benders_annotation(self, group)

    def get_benders_annotation(self):
        return self._model.get_benders_annotation(self)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2019
# --------------------------------------------------------------------------

from docplex.mp.progress import SolutionListener, ProgressClock

from docplex.util.environment import get_environment

import os
import six

try:
    import pandas as pd
except ImportError:
    pd = None

from docplex.util.csv_utils import write_csv, write_table_as_csv

class _KpiRecorder(SolutionListener):
    # '''A specialized subclass of :class:`~SolutionListener` that stores the KPI values
    # at each intermediate solution.
    #
    # The `clock` argument defines which events are listened to.
    #
    # See Also:
    #     the enumerated class :class:`~ProgressClock`
    # '''

    def __init__(self, model, clock=ProgressClock.Gap,
                 publish_hook=None,
                 kpi_publish_format=None, absdiff=None, reldiff=None):
        super(_KpiRecorder, self).__init__(clock, absdiff, reldiff)
        self.model = model
        self._context = model.context
        self.publish_hook = publish_hook
        self.kpi_publish_format = kpi_publish_format or 'KPI.%s'
        self.publish_name_fn = lambda kn: self.kpi_publish_format % kn

        # stored dictionaries of kpi values (name: value)
        self._kpi_dicts = []

    def notify_start(self):
        super(_KpiRecorder, self).notify_start()
        self._kpi_dicts = []

    @property
    def nb_reported(self):
        return len(self._kpi_dicts)

    def notify_solution(self, sol):
        env = get_environment()
        pdata = self.current_progress_data
        context = self._context

        # 1. Start with empty table
        name_values = {}
        # 2. add predefined keys for obj, time.
        name_values['PROGRESS_CURRENT_OBJECTIVE'] = sol.objective_value

        # 3. store it (why???)
        self._kpi_dicts.append(name_values)

        # new stats for https://github.ibm.com/IBMDecisionOptimization/dd-planning/issues/2491
        if env.is_dods():
            name_values['PROGRESS_GAP'] = pdata.mip_gap
            name_values['PROGRESS_BEST_OBJECTIVE'] = pdata.best_bound
            name_values['STAT.cplex.solve.explored'] = pdata.current_nb_nodes
            name_values['STAT.cplex.solve.opened'] = pdata.remaining_nb_nodes
            name_values['STAT.cplex.solve.iterationCount'] = pdata.current_nb_iterations
            name_values['STAT.cplex.solve.elapsedTime'] = pdata.time

        # add KPIs
        publish_name_fn = self.publish_name_fn
        name_values.update({publish_name_fn(kp.name): kp.compute(sol) for kp in self.model.iter_kpis()})
        name_values[publish_name_fn('_time')] = pdata.time

        # usually publish kpis in environment...
        if self.publish_hook is not None:
            self.publish_hook(name_values)

        # save kpis.csv table
        if auto_publishing_kpis_table_names(context) is not None:
            write_kpis_table(env=env,
                             context=context,
                             model=self.model,
                             solution=sol)

    def iter_kpis(self):
        return iter(self._kpi_dicts)

    def __as_df__(self, **kwargs):
        try:
            from pandas import DataFrame
        except ImportError:
            raise RuntimeError("convert as DataFrame: This feature requires pandas")

        df = DataFrame(self._kpi_dicts, **kwargs)
        return df


def get_auto_publish_names(context, prop_name, default_name):
    # comparing auto_publish to boolean values because it can be a non-boolean
    autopubs = context.solver.auto_publish
    if autopubs is True:
        return [default_name]
    elif autopubs is False:
        return None
    elif prop_name in autopubs:
        name = autopubs[prop_name]
    else:
        name = None

    if isinstance(name, six.string_types):
        # only one string value: make this the name of the table
        # in a list with one object
        name = [name]
    elif name is True:
        # if true, then use default name:
        name = [default_name]
    elif name is False:
        # Need to compare explicitely to False
        name = None
    else:
        # otherwise the kpi_table_name can be a collection-like of names,
        # just return it
        pass
    return name


def auto_publishing_result_output_names(context):
    # Return the list of result output names for saving
    return get_auto_publish_names(context, 'result_output', 'solution.json')


def auto_publishing_kpis_table_names(context):
    # Return the list of kpi table names for saving
    return get_auto_publish_names(context, 'kpis_output', 'kpis.csv')


def get_kpis_name_field(context):
    autopubs = context.solver.auto_publish
    field = None
    if autopubs is True:
        field = 'Name'
    elif autopubs is False:
        field = None
    else:
        field = context.solver.auto_publish.kpis_output_field_name
    return field


def get_kpis_value_field(context):
    autopubs = context.solver.auto_publish
    field = None
    if autopubs is True:
        field = 'Value'
    elif autopubs is False:
        field = None
    else:
        field = context.solver.auto_publish.kpis_output_field_value
    return field


class PublishResultAsDf(object):
    '''Mixin for classes publishing a result as data frame
    '''

    @staticmethod
    def value_if_defined(obj, attr_name, default=None):
        value = getattr(obj, attr_name) if hasattr(obj, attr_name) else None
        return value if value is not None else default

    def write_output_table(self, df, context,
                           output_property_name=None,
                           output_name=None):
        '''Publishes the output `df`.

        The `context` is used to control the output name:

            - If context.solver.auto_publish is true, the `df` is written using
            output_name.
            - If context.solver.auto_publish is false, This method does nothing.
            - If context.solver.auto_publish.output_property_name is true,
               then `df` is written using output_name.
            - If context.solver.auto_publish.output_propert_name is None or
            False, this method does nothing.
            - If context.solver.auto_publish.output_propert_name is a string,
            it is used as a name to publish the df

        Example:

            A solver can be defined as publishing a result as data frame::

                class SomeSolver(PublishResultAsDf)
                   def __init__(self, output_customizer):
                      # output something if context.solver.autopublish.somesolver_output is set
                      self.output_table_property_name = 'somesolver_output'
                      # output filename unless specified by somesolver_output:
                      self.default_output_table_name = 'somesolver.csv'
                      # customizer if users wants one
                      self.output_table_customizer = output_customizer
                      # uses pandas.DataFrame if possible, otherwise will use namedtuples
                      self.output_table_using_df = True

                    def solve(self):
                        # do something here and return a result as a df
                        result = pandas.DataFrame(columns=['A','B','C'])
                        return result

            Example usage::

               solver = SomeSolver()
               results = solver.solve()
               solver.write_output_table(results)

        '''

        prop = self.value_if_defined(self, 'output_table_property_name')
        prop = output_property_name if output_property_name else prop
        default_name = self.value_if_defined(self, 'default_output_table_name')
        default_name = output_name if output_name else default_name
        names = get_auto_publish_names(context, prop, default_name)
        use_df = self.value_if_defined(self, 'output_table_using_df', True)
        if names:
            env = get_environment()
            customizer = self.value_if_defined(self, 'output_table_customizer', lambda x: x)
            for name in names:
                r = customizer(df)
                if pd and use_df:
                    env.write_df(r, name)
                else:
                    # assume r is a namedtuple
                    write_csv(env, r, r[0]._fields, name)

    def is_publishing_output_table(self, context):
        prop = self.value_if_defined(self, 'output_table_property_name')
        default_name = self.value_if_defined(self, 'default_output_table_name')
        names = get_auto_publish_names(context, prop, default_name)
        return names


def write_kpis_table(env, context, model, solution):
    names = auto_publishing_kpis_table_names(context)
    kpis_table = []
    for k in model.iter_kpis():
        kpis_table.append([k.name, k.compute(solution)])
    if kpis_table:
        # do not create the kpi tables if there are no kpis to be written
        field_names = [get_kpis_name_field(context),
                       get_kpis_value_field(context)]
        for name in names:
            write_table_as_csv(env, kpis_table, name, field_names)


def write_solution(env, solution, name):
    with env.get_output_stream(name) as output:
        output.write(solution.export_as_json_string().encode('utf-8'))


def write_result_output(env, context, model, solution):
    names = auto_publishing_result_output_names(context)
    for name in names:
        write_solution(env, solution, name)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2018
# --------------------------------------------------------------------------

# gendoc: ignore


def make_solution(col_values, col_names, keep_zeros):
    # Return values as-is if not None, converting all values to float
    if col_values:
        return [float(i) for i in col_values]
    else:
        return []


# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
Tokenizer for reading LP file format
"""

from docplex.cp.tokenizer import *


#==============================================================================
#  Constants
#==============================================================================

_VARIABLE_START_CHARS = set("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&(),;?@_'{}~")
_VARIABLE_BODY_CHARS  = set("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&(),;?@_'{}~0123456789.")

# Section start tokens
TOKEN_KEYWORD_MINIMIZE        = Token(TOKEN_TYPE_KEYWORD, "Minimize")
TOKEN_KEYWORD_MAXIMIZE        = Token(TOKEN_TYPE_KEYWORD, "Maximize")
TOKEN_KEYWORD_MINIMIZE_MULTI  = Token(TOKEN_TYPE_KEYWORD, "Minimize multi-objectives")
TOKEN_KEYWORD_MAXIMIZE_MULTI  = Token(TOKEN_TYPE_KEYWORD, "Maximize multi-objectives")
TOKEN_KEYWORD_SUBJECT_TO      = Token(TOKEN_TYPE_KEYWORD, "Subject to")
TOKEN_KEYWORD_BOUNDS          = Token(TOKEN_TYPE_KEYWORD, "Bounds")
TOKEN_KEYWORD_BINARY          = Token(TOKEN_TYPE_KEYWORD, "Binary")
TOKEN_KEYWORD_GENERAL         = Token(TOKEN_TYPE_KEYWORD, "General")
TOKEN_KEYWORD_NAMELIST        = Token(TOKEN_TYPE_KEYWORD, "Namelist")
TOKEN_KEYWORD_SEMI_CONTINUOUS = Token(TOKEN_TYPE_KEYWORD, "Semi-Continuous")
TOKEN_KEYWORD_SOS             = Token(TOKEN_TYPE_KEYWORD, "SOS")
TOKEN_KEYWORD_PWL             = Token(TOKEN_TYPE_KEYWORD, "PWL")
TOKEN_KEYWORD_END             = Token(TOKEN_TYPE_KEYWORD, "End")

# Map of different wording for sections. Key is section name in lower case, value is token.
_SECTION_NAME_MAP = {
    'minimize': TOKEN_KEYWORD_MINIMIZE, 'minimum':TOKEN_KEYWORD_MINIMIZE, 'min':TOKEN_KEYWORD_MINIMIZE,
    'maximize': TOKEN_KEYWORD_MAXIMIZE, 'maximum': TOKEN_KEYWORD_MAXIMIZE, 'max': TOKEN_KEYWORD_MAXIMIZE,
    'minimize multi-objectives': TOKEN_KEYWORD_MINIMIZE_MULTI,
    'maximize multi-objectives': TOKEN_KEYWORD_MAXIMIZE_MULTI,
    'subject to': TOKEN_KEYWORD_SUBJECT_TO, 'such that': TOKEN_KEYWORD_SUBJECT_TO, 'st': TOKEN_KEYWORD_SUBJECT_TO, 's.t.': TOKEN_KEYWORD_SUBJECT_TO, 'st.': TOKEN_KEYWORD_SUBJECT_TO,
    'user cuts': TOKEN_KEYWORD_SUBJECT_TO, 'lazy constraints': TOKEN_KEYWORD_SUBJECT_TO, # Not documented
    'bounds': TOKEN_KEYWORD_BOUNDS, 'bound': TOKEN_KEYWORD_BOUNDS,
    'binary': TOKEN_KEYWORD_BINARY, 'binaries': TOKEN_KEYWORD_BINARY, 'bin': TOKEN_KEYWORD_BINARY,
    'general': TOKEN_KEYWORD_GENERAL, 'generals': TOKEN_KEYWORD_GENERAL, 'gen': TOKEN_KEYWORD_GENERAL,
    'namelist': TOKEN_KEYWORD_NAMELIST,
    'integers': TOKEN_KEYWORD_GENERAL, 'integer': TOKEN_KEYWORD_GENERAL,  # Not documented, but present in examples
    'semi-continuous': TOKEN_KEYWORD_SEMI_CONTINUOUS, 'semi': TOKEN_KEYWORD_SEMI_CONTINUOUS, 'semis': TOKEN_KEYWORD_SEMI_CONTINUOUS,
    'sos': TOKEN_KEYWORD_SOS,
    'pwl': TOKEN_KEYWORD_PWL,
    'end': TOKEN_KEYWORD_END
}

# Specific tokens
TOKEN_IMPLY       = Token(TOKEN_TYPE_OPERATOR, '->')
TOKEN_IMPLY_BACK  = Token(TOKEN_TYPE_OPERATOR, '<-')
TOKEN_EQUIV       = Token(TOKEN_TYPE_OPERATOR, '<->')
TOKEN_DOUBLECOLON = Token(TOKEN_TYPE_PUNCTUATION, "::")


#==============================================================================
#  Public classes
#==============================================================================

class LpTokenizer(Tokenizer):
    """ Tokenizer for LP file format """
    __slots__ = ()

    def __init__(self, **args):
        """ Create a new tokenizer
        Args:
            See arguments list of :class:`~docplex.cp.utils.Tokenizer`
        """
        super(LpTokenizer, self).__init__(**args)

        # Clear all character handlers
        self.clear_all_char_handlers()

        # Add read of comments
        self.add_char_handler('\\', self._read_antislash)

        # Add symbol read handlers
        for c in _VARIABLE_START_CHARS:
            self.add_char_handler(c, self._read_symbol)

        # Add reading of numbers
        for c in "0123456789":
            self.add_char_handler(c, self._read_number)

        # Add character handlers
        self.add_char_handler('+', lambda: TOKEN_PLUS)
        self.add_char_handler('-', self._read_minus)
        self.add_char_handler('*', lambda: TOKEN_STAR)
        self.add_char_handler('/', lambda: TOKEN_SLASH)
        self.add_char_handler('^', lambda: TOKEN_CIRCUMFLEX)
        self.add_char_handler(':', self._read_colon)
        self.add_char_handler('=', self._read_equal)
        self.add_char_handler('[', lambda: TOKEN_HOOK_OPEN)
        self.add_char_handler(']', lambda: TOKEN_HOOK_CLOSE)
        self.add_char_handler('.', self._read_dot)

        # Add comparison handlers
        self.add_char_handler('>', self._read_greater)
        self.add_char_handler('<', self._read_lower)


    def set_pwl_mode(self, pwl):
        """ Set the tokenizer to parse PWL breakpoints

        Args:
            pwl:  Set PWL mode if true, or restore normal mode if false
        """
        if pwl:
            self.add_char_handler('(', lambda: TOKEN_PARENT_OPEN)
            self.add_char_handler(')', lambda: TOKEN_PARENT_CLOSE)
            self.add_char_handler(',', lambda: TOKEN_COMMA)
        else:
            for c in ('(', ')', ','):
                self.add_char_handler(c, self._read_symbol)

    def _read_minus(self):
        """ Read token starting by - """
        c = self._peek_char()
        if c == '>':
            self._skip_char()
            return TOKEN_IMPLY
        return TOKEN_MINUS


    def _read_equal(self):
        """ Read token starting by = """
        c = self._peek_char()
        if c == '>':
            self._skip_char()
            return TOKEN_GREATER_EQUAL
        if c == '<':
            self._skip_char()
            return TOKEN_LOWER_EQUAL
        return TOKEN_ASSIGN


    def _read_greater(self):
        """ Read token starting by > """
        if self._peek_char() == '=':
            self._skip_char()
            return TOKEN_GREATER_EQUAL
        return TOKEN_GREATER_EQUAL


    def _read_lower(self):
        """ Read token starting by < """
        c = self._peek_char()
        if c == '=':
            self._skip_char()
            return TOKEN_LOWER_EQUAL
        if c == '-':
            self._skip_char()
            if self._peek_char() == '>':
                self._skip_char()
                return TOKEN_EQUIV
            return TOKEN_IMPLY_BACK
        return TOKEN_LOWER_EQUAL


    def _read_symbol(self):
        """ Read and return next symbol """
        # Check section names
        if self.first_in_line:
            # Check undocumented comment
            if self.current_char == '%':
                self._skip_to_end_of_line()
                return Token(TOKEN_TYPE_COMMENT, None if self.skip_comments else self._get_token()[:-1])
            else:
                sname = self.line.strip(' \r\n\t').lower()
                tok = _SECTION_NAME_MAP.get(sname)
                if tok is not None:
                    # Skip all line
                    self.read_index = self.line_length
                    return tok
        c = self._current_char()
        while True:
            if c == '(':
                # Skip all characters up to ')'
                while c and (c != ')'):
                    c = self._next_char()
            c = self._peek_char()
            if c and ((c in _VARIABLE_BODY_CHARS) or (ord(c) > 127)):
                self._next_char()
            else:
                break

        # Check again section name for case where it begins a line that also contains an expression
        tval = self._get_token()
        if self.first_in_line and self._peek_char() <= ' ':
            tok = _SECTION_NAME_MAP.get(tval.lower())
            if tok is not None:
                return tok
        return Token(TOKEN_TYPE_SYMBOL, tval)


    def _read_antislash(self):
        """ Read token starting by \ """
        self._skip_to_end_of_line()
        return Token(TOKEN_TYPE_COMMENT, None if self.skip_comments else self._get_token()[:-1])


    def _read_colon(self):
        """ Read token starting by colon """
        if self._peek_char() == ':':
            self._skip_and_peek_char()
            return TOKEN_DOUBLECOLON
        return TOKEN_COLON


    def _read_dot(self):
        """ Read token starting by dot (particular case for floats) """
        self._back_char()
        return self._read_number()





# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017
# --------------------------------------------------------------------------
# gendoc: ignore

from docplex.mp.utils import is_string

from docplex.mp.sktrans.docplex_mdlr import DOcplexModeler
from docplex.mp.sktrans.cpx_mdlr import CpxModeler


def make_modeler(modeler_name):
    if is_string(modeler_name):
        mkey = modeler_name.lower()
        if mkey == "cplex":
            return CpxModeler()
        elif mkey in {"docplex", "model"}:
            return DOcplexModeler()
        else:
            pass

    raise ValueError("expecting cplex|docplex, {0!r} was passed".format(modeler_name))



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore

import re
import six


class ExchangeFormat(object):
    """ The ExchangeFormat class groups data and attributes of model formats.

        This class is not meant to be instantiated by users.
    """

    def __init__(self, name, extension, requires_cplex, is_binary=False):
        assert isinstance(name, str)
        assert isinstance(extension, str)
        self._name = name
        self._extension = extension if extension.startswith(".") else ".%s" % extension
        self._requires_cplex = bool(requires_cplex)
        self._is_binary = is_binary

    @property
    def filetype(self):
        # remove starting '.' from extension
        return self._extension[1:]

    @property
    def name(self):
        """ Returns a string qualifying the format to be used in messages.

        Example:
            LP format name is "LP"
        """
        return self._name

    @property
    def is_binary(self):
        """ Returns True if the format is binary, False for a text format.

        Example:
            LP is a text format and returns False.

            SAV is a binary format and returns TRue
        """
        return self._is_binary

    @property
    def extension(self):
        """ Returns the full file extension of the format, including "."

        Example:
            LP format extension is ".lp". SAV format extension is ".sav"
        """
        return self._extension

    def to_string(self):
        return "%s_format" % self.name

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()

    @staticmethod
    def fromstring(definition):
        if definition:
            definition = definition.lower()
        return _FORMAT_MAPPER.get(definition, None)


class LPFormat(ExchangeFormat):

    __raw = " -+/\\<>"
    __cooked = "_mp____"

    from docplex.mp.compat23 import mktrans

    _str_translate_table = mktrans(__raw, __cooked)
    _unicode_translate_table = {}
    for c in range(len(__raw)):
        _unicode_translate_table[ord(__raw[c])] = ord(__cooked[c])

    @classmethod
    def _translate_chars_py2(cls, raw_name):
        # noinspection PyUnresolvedReferences
        if isinstance(raw_name, unicode):
            char_mapping = cls._unicode_translate_table
        else:
            char_mapping = cls._str_translate_table
        return raw_name.translate(char_mapping)

    @classmethod
    def _translate_chars_py3(cls, raw_name):
        return raw_name.translate(cls._unicode_translate_table)

    # which translate_method to use
    if six.PY2:
        _translate_chars = _translate_chars_py2
    else:  # pragma: no cover
        _translate_chars = _translate_chars_py3

    def __init__(self):
        ExchangeFormat.__init__(self, "LP", "lp", requires_cplex=False)

    lp_re = re.compile(r"[a-df-zA-DF-Z!#$%&()/,;?@_`'{}|\"][a-zA-Z0-9!#$%&()/.,;?@_`'{}|\"]*")

    @classmethod
    def is_lp_compliant(cls, name, lp_re=lp_re):
        if name:
            lp_match = lp_re.match(name)
            return lp_match and lp_match.start() == 0 and lp_match.end() == len(name)
        else:
            return False

    @classmethod
    def make_prefix_name(cls, prefix, local_index, offset=1):
        prefixed_name = "{0:s}{1:d}".format(prefix, local_index + offset)
        return prefixed_name

    @classmethod
    def lp_var_name(cls, dvar):
        return cls.lp_name(dvar.get_name(), "x", dvar.get_index())

    # @classmethod
    # def lp_ct_name(cls, linct):
    #     return cls.lp_name(linct.get_name(), "c", linct.get_index())

    @classmethod
    def lp_name(cls, raw_name, prefix, local_index, hide_names=False,
                noncompliant_hook=None):
        # anonymous constraints must be named in a LP (we follow CPLEX here)
        if hide_names or not raw_name:
            return cls.make_prefix_name(prefix, local_index, offset=1)

        # # swap blanks with underscores
        ws_name = cls._translate_chars(raw_name)

        if not cls.is_lp_compliant(ws_name):
            if ws_name[0] in 'eE':
                # fixing eE non-LP names
                fixed_name = '_' + ws_name
                if cls.is_lp_compliant(fixed_name):
                    return fixed_name
            # -- stats
            if noncompliant_hook:
                noncompliant_hook(raw_name)
            # --
            return cls.make_prefix_name(prefix, local_index, offset=1)
        else:
            # truncate if necessary, again this does nothing if name is too short
            return ws_name[:255]


""" The global LP format object."""
# noinspection PyPep8
LP_format  = LPFormat()
SAV_format = ExchangeFormat("SAV", "sav", requires_cplex=True, is_binary=True)
OPL_format = ExchangeFormat("OPL", ".mod", requires_cplex=False)
MPS_format = ExchangeFormat("MPS", ".mps", requires_cplex=True)

_FORMAT_MAPPER = {"lp": LP_format,
                  "opl": OPL_format
                  }



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

from docplex.mp.operand import Operand
from docplex.mp.error_handler import docplex_fatal, DOcplexException

from docplex.mp.utils import is_number, is_string, is_function, str_maxed


class KPI(object):
    """ Abstract class for key performance indicators (KPIs).

    Each KPI has a unique name. A KPI is attached to a model instance and can compute a numerical value,
    using the :func:`compute` method. The `compute` method takes an optional solution argument;
    if passed a valid SolveSolution object, this solution is used to evaluate the KPI, else compute()
    will attempt to access th emodel's solution. If the model has no attached solution, then an exception
    is raised by `compute`.

    """

    def __init__(self):
        pass

    def get_name(self):
        """
        Returns:
            string: The published name of the KPI, a non-empty, unique string.
        """
        raise NotImplementedError  # pragma: no cover

    def set_name(self, new_name):
        pass  # pragma: no cover

    def get_model(self):
        """
        Returns:
           The model instance on which the KPI is defined.
        :rtype: :class:`docplex.mp.model.Model`
        """
        raise NotImplementedError   # pragma: no cover

    def compute(self, s=None):
        raise NotImplementedError   # pragma: no cover

    def _get_solution_value(self, s=None):  # pragma: no cover
        return self.compute(s)

    def _ensure_solution(self, s, do_raise=True):
        # INTERNAL
        if s is not None:
            return s
        else:
            ms = self.get_model()._solution
            if ms is not None:
                return ms
            elif do_raise:
                self.get_model().fatal("KPI.compute() requires a solution, but model is not solved and no solution was passed")
            else:
                return None

    @property
    def solution_value(self):
        return self._get_solution_value()

    def _check_name(self, name_arg):
        if not is_string(name_arg) or not name_arg:
            self.get_model().fatal("KPI.set_name() expects a non-empty string, got: {0!r}", name_arg)

    def is_decision_expression(self):
        """ returns True if the KPI is based on a decision expression or variable.

        """
        raise NotImplementedError   # pragma: no cover

    def copy(self, new_model, var_map):
        raise NotImplementedError   # pragma: no cover

    def clone(self):
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def new_kpi(model, kpi_arg, kpi_name):
        # static factory method to build a new concrete instance of KPI
        if isinstance(kpi_arg, KPI):
            if not kpi_name:
                return kpi_arg
            else:
                cloned = kpi_arg.clone()
                cloned.name = kpi_name
                return cloned
        elif is_function(kpi_arg):
            return FunctionalKPI(kpi_arg, model, kpi_name)
        else:
            # try a linear expr conversion
            try:
                expr = model._lfactory._to_expr(kpi_arg)
                return DecisionKPI(expr, kpi_name)
            except DOcplexException:
                model.fatal("Cannot interpret this as a KPI: {0!r}. expecting expression, variable or function", kpi_arg)

    def notify_removed(self):
        pass


class DecisionKPI(KPI):
    """ Specialized class of Key Performance Indicator, based on expressions.

    This subclass is built from a decision variable or a linear expression.
    The :func:`compute` method evaluate the value of the KPI in a solution. This solution can either be passed
    to the `compute` method, or using th emodel's solution. In the latter case, the model must have been solved
    with a solution.

    """
    def __init__(self, kpi_op, name=None):
        KPI.__init__(self)
        if is_number(kpi_op):
            self._expr = self.get_model().linear_expr(arg=kpi_op)
            self._name = name
        elif not isinstance(kpi_op, Operand):
            self.get_model().fatal('cannot interpret this as kpi: {0!r}', kpi_op)
        else:
            self._expr = kpi_op
            self._expr.notify_used(self)  # kpi is a subscriber
            self._name = name or kpi_op.name
        self._check_name(self._name)

    def notify_expr_modified(self, expr, event):
        # do nothing
        pass

    def notify_removed(self):
        self._expr.notify_unsubscribed(self)

    def get_name(self):
        return self._name

    def set_name(self, new_name):
        self._check_name(new_name)
        self._name = new_name

    name = property(get_name, set_name)

    def get_model(self):
        return self._expr.model

    def compute(self, s=None):
        """ Redefinition of the abstract `compute(s=None)` method.

        Returns:
            float: The value of the decision expression in the solution.

        Note:
            Evaluating a KPI requires a solution object. This solution can either be passed explicitly
            in the `s` argument, otherwise the model solution is used. In the latter case, th emodel must
            have been solved with a solution, otherwise an exception is raised.

        Raises:
            Evaluating a KPI raises an exception if no `s` solution has been passed
            and the underlying model has not been solved with a solution.

        See Also:
            :class:`docplex.mp.solution.SolveSolution`
        """
        es = self._ensure_solution(s, do_raise=True)
        return self._expr._get_solution_value(es)

    def is_decision_expression(self):
        return True

    def to_expr(self):
        return self._expr

    as_expression = to_expr

    def to_linear_expr(self):
        return self._expr.to_linear_expr()

    def copy(self, new_model, var_map):
        expr_copy = self._expr.copy(new_model, var_map)
        return DecisionKPI(kpi_op=expr_copy, name=self.name)

    def clone(self):
        return DecisionKPI(self._expr, self._name)

    def __repr__(self):
        return "{0}(name={1},expr={2!s})".format(self.__class__.__name__, self.name, str_maxed(self._expr, maxlen=64))


class FunctionalKPI(KPI):
    # Functional KPIs store a function that takes a model to compute a number
    # Functional KPIs do not require a successful solve.

    def __init__(self, fn, model, name):
        KPI.__init__(self)
        self._name = name
        self._function = fn
        self._model = model
        self._check_name(self._name)

    def get_name(self):
        return self._name

    def set_name(self, new_name):
        self._name = new_name

    name = property(get_name, set_name)

    def get_model(self):
        return self._model

    def compute(self, s=None):
        es = self._ensure_solution(s)
        return self._function(self._model, es)

    def is_decision_expression(self):
        return False

    def copy(self, new_model, var_map):
        return FunctionalKPI(fn=self._function, model=new_model, name=self._name)

    def clone(self):
        return FunctionalKPI(fn=self._function, model=self._model, name=self._name)

    def to_expr(self):
        docplex_fatal("This KPI cannot be used as an expression: {0!r}".format(self))



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2018
# --------------------------------------------------------------------------

try:
    from pyspark import keyword_only  ## < 2.0 -> pyspark.ml.util.keyword_only
    from pyspark.ml import Transformer
    from pyspark.ml.param.shared import Param

    from pyspark.sql import SparkSession
    spark_version = SparkSession.builder.getOrCreate().version
except ImportError:
    spark_version = None
    def keyword_only(x):
        return x 
    Transformer = object
    Param = SparkSession = None

from docplex.mp.constants import ObjectiveSense
from docplex.mp.sktrans.modeler import make_modeler
from docplex.mp.utils import *
from docplex.mp.sparktrans.spark_utils import make_solution


def convert_to_list_or_value(value):
    if is_spark_dataframe(value):
        return value.rdd.flatMap(lambda x: x).collect()
    elif is_pandas_series(value):
        return value.tolist()
    elif is_numpy_ndarray(value):
        return value.tolist()
    return value


class CplexTransformerBase(Transformer):
    """ Root class for CPLEX transformers for PySpark
    """

    @keyword_only
    def __init__(self, rhsCol=None, minCol=None, maxCol=None, y=None, lbs=None, ubs=None, types=None, sense="min", modeler="cplex", solveParams=None):
        super(CplexTransformerBase, self).__init__()
        self.rhsCol = Param(self, "rhsCol", "")
        self.minCol = Param(self, "minCol", "")
        self.maxCol = Param(self, "maxCol", "")
        self.y = Param(self, "y", "")
        self.lbs = Param(self, "lbs", "")
        self.ubs = Param(self, "ubs", "")
        self.types = Param(self, "types", "")
        self.sense = Param(self, "sense", "")
        self.modeler = Param(self, "modeler", "")
        self.solveParams = Param(self, "solveParams", "")
        self._setDefault(rhsCol=None)
        self._setDefault(minCol=None)
        self._setDefault(maxCol=None)
        self._setDefault(y=None)
        self._setDefault(lbs=None)
        self._setDefault(ubs=None)
        self._setDefault(types=None)
        self._setDefault(sense=sense)
        self._setDefault(modeler=modeler)
        self._setDefault(solveParams={})

        if spark_version < '2.2':
            kwargs = self.__init__._input_kwargs
        else:
            kwargs = self._input_kwargs
        self.setParams(**kwargs)

        self.docplex_modeler = make_modeler(self.getModeler())

    @keyword_only
    def setParams(self, rhsCol=None, minCol=None, maxCol=None, y=None, lbs=None, ubs=None, types=None, solveParams=None, sense=None, modeler=None):
        if spark_version < '2.2':
            kwargs = self.__init__._input_kwargs
        else:
            kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setRhsCol(self, value):
        self._paramMap[self.rhsCol] = value
        return self

    def getRhsCol(self):
        return self.getOrDefault(self.rhsCol)

    def setMinCol(self, value):
        self._paramMap[self.minCol] = value
        return self

    def getMinCol(self):
        return self.getOrDefault(self.minCol)

    def setMaxCol(self, value):
        self._paramMap[self.maxCol] = value
        return self

    def getMaxCol(self):
        return self.getOrDefault(self.maxCol)

    def setY(self, value):
        self._paramMap[self.y] = value
        return self

    def getY(self):
        return self.getOrDefault(self.y)

    def setLbs(self, value):
        self._paramMap[self.lbs] = value
        return self

    def getLbs(self):
        return self.getOrDefault(self.lbs)

    def setUbs(self, value):
        self._paramMap[self.ubs] = value
        return self

    def getUbs(self):
        return self.getOrDefault(self.ubs)

    def setTypes(self, value):
        self._paramMap[self.types] = value
        return self

    def getTypes(self):
        return self.getOrDefault(self.types)

    def setSense(self, value):
        self._paramMap[self.sense] = value
        return self

    def getSense(self):
        return self.getOrDefault(self.sense)

    def setModeler(self, value):
        self._paramMap[self.modeler] = value
        return self

    def getModeler(self):
        return self.getOrDefault(self.modeler)

    def setSolveParams(self, value):
        self._paramMap[self.solveParams] = value
        return self

    def getSolveParams(self):
        return self.getOrDefault(self.solveParams)

    def _transform(self, dataset):
        """ Main method to solve Linear Programming problems.
        Transforms the input dataset.

        :param dataset: the matrix describing the  constraints of the problem, which is an instance
                of :py:class:`pyspark.sql.DataFrame`
        :returns: transformed dataset
        """
        objSense = ObjectiveSense.parse(self.getSense())

        var_lbs = convert_to_list_or_value(self.getLbs())
        var_ubs = convert_to_list_or_value(self.getUbs())
        var_types = self.getTypes()

        min_max_cols = [self.getMinCol(), self.getMaxCol()] if self.getMinCol() else [self.getRhsCol()]
        coef_colnames = [item for item in dataset.columns if item not in min_max_cols]
        nb_vars = len(coef_colnames)

        # Get min and max values as lists
        minVals, maxVals = self._get_min_max_lists(dataset)

        # Get coefficients for objective evaluation
        costs = None if self.getY() is None else convert_to_list_or_value(self.getY())
        assert (costs is None) or (len(costs) == nb_vars)

        # Build matrix of coefficients
        cts_mat = dataset.rdd.map(lambda x: [x[col] for col in coef_colnames]).collect()

        # Call build matrix and solve
        params = self.getSolveParams()  # User-defined engine parameters
        if minVals is None:
            result = self.docplex_modeler.build_matrix_linear_model_and_solve(nb_vars, var_lbs, var_ubs, var_types,
                                                                              coef_colnames,
                                                                              cts_mat, maxVals,
                                                                              objsense=objSense, costs=costs,
                                                                              cast_to_float=True,
                                                                              solution_maker=make_solution,
                                                                              **params)
        else:
            result = self.docplex_modeler.build_matrix_range_model_and_solve(nb_vars, var_lbs, var_ubs, var_types, coef_colnames,
                                                                             cts_mat=cts_mat,
                                                                             range_mins=minVals, range_maxs=maxVals,
                                                                             objsense=objSense, costs=costs,
                                                                             cast_to_float=True,
                                                                             solution_maker=make_solution,
                                                                             **params)
        # 'result' is a list: create a Spark DataFrame from it
        sparkCtx = dataset.sql_ctx
        return sparkCtx.createDataFrame(zip(coef_colnames, result), schema=['name', 'value'])


class CplexTransformer(CplexTransformerBase):
    """ A PySpark transformer class to solve linear problems.

    This transformer class solves LP problems of
    type::

            Ax <= B

    """

    @keyword_only
    def __init__(self, rhsCol=None, y=None, lbs=None, ubs=None, types=None, sense="min", modeler="cplex", solveParams=None):
        """
        Creates an instance of LPTransformer to solve linear problems.

        :param rhsCol: the name of the column in the input Spark DataFrame containing the upper bounds for the constraints.
        :param y: an optional sequence of scalars describing the cost vector
        :param lbs: an optional sequence of scalars describing the lower bounds for decision variables
        :param ubs: an optional sequence of scalars describing the upper bounds for decision variables
        :param types: a string for variable types within [BICSN]*
        :param sense: defines the objective sense. Accepts 'min" or "max" (not case-sensitive),
            or an instance of docplex.mp.ObjectiveSense
        :param solveParams: optional keyword arguments to pass additional parameters for Cplex engine.

        Note:
            The Spark dataframe representing the matrix is supposed to have shape (M,N+1) where M is the number of rows
            and N the number of variables. The column named by the 'rhsCol' parameter contains the right hand sides of the problem
            (the B in Ax <= B)
            The optional vector y contains the N cost coefficients for each column variables.

        Example:
            Passing the Spark DataFrame = ([[1,2,3], [4,5,6]], ['x', 'y', 'max']), rhsCol = 'max', y= [11,12] means
            solving the linear problem:

                minimize 11x + 12y
                s.t.
                        1x + 2y <= 3
                        4x + 5y <= 6
        """
        if spark_version < '2.2':
            super(CplexTransformer, self).__init__(**self.__init__._input_kwargs)
        else:
            super(CplexTransformer, self).__init__(**self._input_kwargs)

    def _get_min_max_lists(self, dataset):
        assert self.getRhsCol() and self.getRhsCol() in dataset.columns, '"rhsCol" parameter is not specified'

        maxVals = dataset.select(self.getRhsCol()).rdd.flatMap(lambda x: x).collect()
        return None, maxVals


class CplexRangeTransformer(CplexTransformerBase):
    """ A PySpark transformer class to solve range-based linear problems.

    This transformer class solves LP problems of
    type::

            B <= Ax <= C

    """

    @keyword_only
    def __init__(self, minCol=None, maxCol=None, y=None, lbs=None, ubs=None, types=None, sense="min", modeler="cplex", solveParams=None):
        """
        Creates an instance of LPRangeTransformer to solve range-based linear problems.

        :param minCol: the name of the column in the input Spark DataFrame containing the lower bounds for the constraints.
        :param maxCol: the name of the column in the input Spark DataFrame containing the upper bounds for the constraints.
        :param y: an optional sequence of scalars describing the cost vector
        :param lbs: an optional sequence of scalars describing the lower bounds for decision variables
        :param ubs: an optional sequence of scalars describing the upper bounds for decision variables
        :param types: a string for variable types within [BICSN]*
        :param sense: defines the objective sense. Accepts 'min" or "max" (not case-sensitive),
            or an instance of docplex.mp.ObjectiveSense
        :param solveParams: optional keyword arguments to pass additional parameters for Cplex engine.

        Note:
            The matrix X is supposed to have shape (M,N+2) where M is the number of rows
            and N the number of variables.
            The column named by the 'minCol' and 'maxCol' parameters contain the minimum (resp.maximum) values for the
            row ranges, that m and M in:
                    m <= Ax <= M
            The optional vector y contains the N cost coefficients for each column variables.

        Example:
            Passing the Spark DataFrame = ([[1,2,3,30], [4,5,6,60]], ['x', 'y', 'min', 'max']), minCol = 'min',
            maxCol = 'max', y= [11,12] means solving the linear problem:

                minimize 11x + 12y
                s.t.
                        3 <= 1x + 2y <= 30
                        6 <= 4x + 5y <= 60
        """
        if spark_version < '2.2':
            super(CplexRangeTransformer, self).__init__(**self.__init__._input_kwargs)
        else:
            super(CplexRangeTransformer, self).__init__(**self._input_kwargs)

    def _get_min_max_lists(self, dataset):
        assert self.getMinCol() and self.getMinCol() in dataset.columns, '"minCol" parameter is not specified'
        assert self.getMaxCol() and self.getMaxCol() in dataset.columns, '"maxCol" parameter is not specified'

        minVals = dataset.select(self.getMinCol()).rdd.flatMap(lambda x: x).collect()
        maxVals = dataset.select(self.getMaxCol()).rdd.flatMap(lambda x: x).collect()
        return minVals, maxVals



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
Tokenizer for reading CPO file format
"""

from docplex.cp.tokenizer import *


#==============================================================================
#  Constants
#==============================================================================

TOKEN_IMPLY     = Token(TOKEN_TYPE_OPERATOR, '=>')
TOKEN_AND       = Token(TOKEN_TYPE_OPERATOR, '&&')
TOKEN_OR        = Token(TOKEN_TYPE_OPERATOR, '||')
TOKEN_NOT       = Token(TOKEN_TYPE_OPERATOR, '!')
TOKEN_INTERVAL  = Token(TOKEN_TYPE_OPERATOR, '..')
TOKEN_DIV       = Token(TOKEN_TYPE_OPERATOR, 'div')
TOKEN_LET       = Token(TOKEN_TYPE_SYMBOL,   'let')
TOKEN_SET       = Token(TOKEN_TYPE_SYMBOL,   'set')


#==============================================================================
#  Public classes
#==============================================================================

class CpoTokenizer(Tokenizer):
    """ Tokenizer for CPO file format """
    __slots__ = ()

    def __init__(self, **args):
        """ Create a new tokenizer
        Args:
            See arguments list of :class:`~docplex.cp.utils.Tokenizer`
        """
        super(CpoTokenizer, self).__init__(**args)

        # Add predefined symbols
        self.symbols = {}
        self.add_predefined_symbols([TOKEN_DIV])
        self.add_predefined_symbols([TOKEN_LET])
        self.add_predefined_symbols([TOKEN_SET])

        # Add character handlers
        self.add_char_handler('.', self._read_dot)
        self.add_char_handler('=', self._read_equal)
        self.add_char_handler('!', self._read_bang)
        self.add_char_handler('&', self._read_and)
        self.add_char_handler('|', self._read_pipe)
        self.add_char_handler('/', self._read_slash)


    def _read_dot(self):
        """ Read token starting by dot """
        if self._next_char() != '.':
            raise SyntaxError(self.build_error_string("Unknown token '.'"))
        return TOKEN_INTERVAL


    def _read_equal(self):
        """ Read token starting by - """
        c = self._peek_char()
        if c == '=':
            self._skip_char()
            return TOKEN_EQUAL
        elif c == '>':
            self._skip_char()
            return TOKEN_IMPLY
        return TOKEN_ASSIGN


    def _read_bang(self):
        """ Read token starting by ! """
        if self._peek_char() == '=':
            self._skip_char()
            return TOKEN_DIFFERENT
        return TOKEN_NOT


    def _read_and(self):
        """ Read token starting by & """
        if self._next_char() != '&':
            raise SyntaxError(self.build_error_string("Unknown token '&'"))
        return TOKEN_AND


    def _read_pipe(self):
        """ Read token starting by | """
        if self._next_char() != '|':
            raise SyntaxError(self.build_error_string("Unknown token '|'"))
        return TOKEN_OR


    def _read_slash(self):
        """ Read token starting by / """
        c = self._peek_char()
        if c == '/':
            self._skip_to_end_of_line()
            return Token(TOKEN_TYPE_COMMENT, None if self.skip_comments else self._get_token()[:-1])
        elif c == '*':
            # Read comment
            cbuff = None if self.skip_comments else ['/', '*']
            self._skip_char()
            pc = ''
            c = self._next_char()
            if cbuff:
                cbuff.append(c)
            while c and ((c != '/') or (pc != '*')):
                pc = c
                c = self._next_char()
                if cbuff:
                    cbuff.append(c)
            return Token(TOKEN_TYPE_COMMENT, ''.join(cbuff) if cbuff else None)
        return TOKEN_SLASH


